'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var rollupPluginCommonjs = _interopDefault(require('rollup-plugin-commonjs'));
var rollupPluginJson = _interopDefault(require('rollup-plugin-json'));
var rollupPluginReplace = _interopDefault(require('rollup-plugin-replace'));
var fs = _interopDefault(require('fs'));
var path = _interopDefault(require('path'));
var rimraf = _interopDefault(require('rimraf'));
var isModule$1 = _interopDefault(require('is-module'));
var util = _interopDefault(require('util'));
var chalk = _interopDefault(require('chalk'));
var pirates = _interopDefault(require('pirates'));
var glob = _interopDefault(require('glob'));
var module$1 = _interopDefault(require('module'));
var sucrase = _interopDefault(require('sucrase'));
var url = _interopDefault(require('url'));
var rollupPluginNodeResolve = _interopDefault(require('rollup-plugin-node-resolve'));
var rollup$1 = _interopDefault(require('rollup'));

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

const TESTING = process.env.NODE_ENV === 'test';
// Prevent parallel test runs from reading from same cache
const DIR = `.dvlp${TESTING ? process.getuid() : ''}`;
const VERSION = process.env.DVLP_VERSION;

const dir = path.resolve(DIR);
const bundleDirName = `${path.join(DIR, `bundle-${VERSION}`)}`;
const bundleDir = path.resolve(bundleDirName);
const maxModuleBundlerWorkers = parseInt(process.env.BUNDLE_WORKERS, 10) || 0;
const port = process.env.PORT ? Number(process.env.PORT) : 8080;

// Work around rollup-plugin-commonjs require.main.filename
if (TESTING || process.env.DVLP_LAUNCHER === 'cmd') {
  const bundleDirExists = fs.existsSync(bundleDir);
  const dirExists = fs.existsSync(dir);
  const rm = dirExists && !bundleDirExists;

  if (rm) {
    const contents = fs.readdirSync(dir).map((item) => path.resolve(dir, item));

    for (const item of contents) {
      // Delete all subdirectories
      if (fs.statSync(item).isDirectory()) {
        rimraf.sync(item);
      }
    }
  }
  if (!dirExists) {
    fs.mkdirSync(dir);
  }
  if (!bundleDirExists) {
    fs.mkdirSync(bundleDir);
  }

  if (TESTING) {
    process.on('exit', () => {
      rimraf.sync(dir);
    });
  }
}

/**
 * @typedef { object } Config
 * @property { number } activePort,
 * @property { string } bundleDir,
 * @property { string } bundleDirName,
 * @property { Array<string> } directories,
 * @property { object } extensionsByType,
 * @property { Array<string> } extensionsByType.css,
 * @property { Array<string> } extensionsByType.html,
 * @property { Array<string> } extensionsByType.js,
 * @property { number } latency,
 * @property { string } maxAge,
 * @property { number } maxModuleBundlerWorkers,
 * @property { number } port,
 * @property { string } testing,
 * @property { object } typesByExtension,
 * @property { object } typesByExtension,
 */
var config = {
  activePort: port,
  bundleDir,
  bundleDirName,
  directories: [process.cwd()],
  extensionsByType: {
    css: ['.css', '.sass', '.scss', '.less', '.styl', '.stylus'],
    html: [
      '.html',
      '.htm',
      '.nunjs',
      '.nunjucks',
      '.hbs',
      '.handlebars',
      '.dust'
    ],
    js: ['.js', '.mjs', '.jsx', '.ts', '.tsx', '.json']
  },
  latency: 50,
  maxAge: '10m',
  maxModuleBundlerWorkers,
  port,
  rollupConfigPath: path.join(dir, 'rollup.config.js'),
  testing: TESTING,
  typesByExtension: {
    '.sass': 'css',
    '.scss': 'css',
    '.less': 'css',
    '.styl': 'css',
    '.stylus': 'css',
    '.html': 'html',
    '.htm': 'html',
    '.nunjs': 'html',
    '.nunjucks': 'html',
    '.hbs': 'html',
    '.handlebars': 'html',
    '.dust': 'html',
    '.mjs': 'js',
    '.json': 'js',
    '.jsx': 'js',
    '.ts': 'js',
    '.tsx': 'js'
  }
};

/**
 * @typedef { import("http").ClientRequest } ClientRequest
 */







const HAS_UTIL_IS_PROXY = util.types && util.types.isProxy;
const RE_BARE_IMPORT = /^[^./]/;
const RE_INVALID = /[<>:"|?*]/;
const RE_JSON = /.json$/i;
const RE_LOCALHOST = /localhost|127\.0\.0\.1/;
const RE_NODE_MODULES = /node_modules/;
const RE_TYPE_CSS = /text\/css/i;
const RE_TYPE_HTML = /text\/html/i;
const RE_TYPE_JS = /application\/javascript/i;

var is = {
  isAbsoluteFilePath,
  isBareImport,
  isCssFilePath,
  isCssRequest,
  isHtmlFilePath,
  isHtmlRequest,
  isInvalidFilePath,
  isJsFilePath,
  isJsRequest,
  isJsonFilePath,
  isLocalhost,
  isNodeModuleFilePath,
  isModule,
  isModuleBundlerFilePath,
  isProjectFilePath,
  isPromise,
  isProxy,
  isRelativeFilePath,
  isValidFilePath
};

/**
 * Determine if "filePath" is absolute
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isAbsoluteFilePath(filePath) {
  return (
    'string' == typeof filePath &&
    path.isAbsolute(filePath) &&
    // Only absolute if from root
    filePath.startsWith(process.cwd().slice(0, 5))
  );
}

/**
 * Determine if 'id' is referencing a node_module
 *
 * @param { string } id
 * @returns { boolean }
 */
function isBareImport(id) {
  return RE_BARE_IMPORT.test(id);
}

/**
 * Determine if 'filePath' is for a css file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isCssFilePath(filePath) {
  return config.extensionsByType.css.includes(path.extname(filePath));
}

/**
 * Determine if 'req' is for a css file
 *
 * @param { ClientRequest } req
 * @returns { boolean }
 */
function isCssRequest(req) {
  return (
    req.type === 'css' ||
    RE_TYPE_CSS.test(req.headers.accept) ||
    isCssFilePath(req.url)
  );
}

/**
 * Determine if 'filePath' is for an html file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isHtmlFilePath(filePath) {
  return config.extensionsByType.html.includes(path.extname(filePath));
}

/**
 * Determine if 'req' is for an html file
 *
 * @param { ClientRequest } req
 * @returns { boolean }
 */
function isHtmlRequest(req) {
  return (
    req.type === 'html' ||
    RE_TYPE_HTML.test(req.headers.accept) ||
    isHtmlFilePath(req.url)
  );
}

/**
 * Determine if 'filePath' is invalid
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isInvalidFilePath(filePath) {
  return RE_INVALID.test(filePath);
}

/**
 * Determine if 'filePath' is for a js file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isJsFilePath(filePath) {
  return config.extensionsByType.js.includes(path.extname(filePath));
}

/**
 * Determine if 'req' is for a js file
 *
 * @param { ClientRequest } req
 * @returns { boolean }
 */
function isJsRequest(req) {
  return (
    req.type === 'js' ||
    // Almost always '*/*'
    RE_TYPE_JS.test(req.headers.accept) ||
    isJsFilePath(req.url)
  );
}

/**
 * Determine if 'filePath' is for a js file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isJsonFilePath(filePath) {
  return RE_JSON.test(filePath);
}

/**
 * Determine if 'url' is localhost
 *
 * @param { string } url
 * @returns { boolean }
 */
function isLocalhost(url) {
  return RE_LOCALHOST.test(url);
}

/**
 * Determine if 'filePath' is in node_modules
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isNodeModuleFilePath(filePath) {
  return RE_NODE_MODULES.test(filePath);
}

/**
 * Determine if filePath or code is es module
 *
 * @param { string } filePathOrCode
 * @returns { boolean }
 */
function isModule(filePathOrCode) {
  if (isJsFilePath(filePathOrCode)) {
    filePathOrCode = fs.readFileSync(filePathOrCode, 'utf8');
  }
  return isModule$1(filePathOrCode);
}

/**
 * Determine if 'filePath' is for a bundled module file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isModuleBundlerFilePath(filePath) {
  return filePath.includes(config.bundleDirName);
}

/**
 * Determine if 'obj' is Proxy
 *
 * @param { object } obj
 * @returns { boolean }
 */
function isProxy(obj) {
  if (HAS_UTIL_IS_PROXY) {
    return util.types.isProxy(obj);
  } else {
    return obj instanceof Proxy;
  }
}

/**
 * Determine if 'obj' is Promise instance
 *
 * @param { object } obj
 * @returns { boolean }
 */
function isPromise(obj) {
  if (HAS_UTIL_IS_PROXY) {
    return util.types.isPromise(obj);
  } else {
    return obj instanceof Promise;
  }
}

/**
 * Determine if 'filePath' is in project
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isProjectFilePath(filePath) {
  if (!path.isAbsolute(filePath)) {
    filePath = path.resolve(filePath);
  }
  return filePath.includes(process.cwd()) && !isNodeModuleFilePath(filePath);
}

/**
 * Determine if "filePath" is relative
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isRelativeFilePath(filePath) {
  return 'string' == typeof filePath && filePath.startsWith('.');
}

/**
 * Determine if "filePath" is valid.
 * If relative, resolves against "fromFilePath".
 *
 * @param { string } filePath
 * @param { string } [fromDir]
 * @returns { boolean }
 */
function isValidFilePath(filePath, fromDir = process.cwd()) {
  if (isRelativeFilePath(filePath)) {
    filePath = path.join(fromDir, filePath);
  }

  try {
    const stat = fs.statSync(filePath);
    if (stat.isFile()) {
      return true;
    }
  } catch (err) {
    return false;
  }
}

const SEG_LENGTH = 80;
const WARN_BARE_IMPORT =
  '⚠️  re-writing bare imports because browsers do not support Node-style import identifiers';
const WARN_MISSING_EXTENSION =
  '⚠️  adding missing extensions because browsers can only resolve valid URL identifiers';
const WARN_NODE_PATH =
  '⚠️  re-writing NODE_PATH import because browsers can only resolve valid URL identifiers';
const WARN_PACKAGE_INDEX =
  '⚠️  adding missing package index.js because browsers do not support Node-style package identifiers';
const WARN_SERVER_TRANSPILE =
  '⚠️  ignoring async Promise returned when executing transpiler on a server file. Be sure to check the "isServer" argument before returning a transpiled value';

let maxBareImport = 1;
let maxMissingExtension = 1;
let maxNodePath = 1;
let maxPackageIndex = 1;
let maxServerTranspile = 1;
let silent = false;

var log = {
  WARN_BARE_IMPORT,
  WARN_MISSING_EXTENSION,
  WARN_NODE_PATH,
  WARN_PACKAGE_INDEX,
  WARN_SERVER_TRANSPILE,
  info,
  noisyInfo,
  warn,
  error,
  fatal,
  set silent(value) {
    silent = value;
  }
};

/**
 * Log if not testing/silent
 *
 * @param { ...any } args
 */
function info(...args) {
  if (!config.testing && !silent) {
    console.log(' ', ...args.map(truncate));
  }
}

/**
 * Log if not testing, even if silent
 *
 * @param { ...any } args
 */
function noisyInfo(...args) {
  if (!config.testing) {
    console.log(' ', ...args.map(truncate));
  }
}

/**
 * Warn if not testing/silent
 *
 * @param { ...any } args
 */
function warn(...args) {
  if (!config.testing) {
    const [warning] = args;
    let msg = '';

    switch (warning) {
      case WARN_BARE_IMPORT:
        if (maxBareImport-- > 0) {
          msg = `${WARN_BARE_IMPORT}`;
        }
        break;
      case WARN_NODE_PATH:
        if (maxNodePath-- > 0) {
          msg = `${WARN_NODE_PATH}`;
        }
        break;
      case WARN_MISSING_EXTENSION:
        if (maxMissingExtension-- > 0) {
          msg = `${WARN_MISSING_EXTENSION}`;
        }
        break;
      case WARN_PACKAGE_INDEX:
        if (maxPackageIndex-- > 0) {
          msg = `${WARN_PACKAGE_INDEX}`;
        }
        break;
      case WARN_SERVER_TRANSPILE:
        if (maxServerTranspile-- > 0) {
          msg = `${WARN_SERVER_TRANSPILE}`;
        }
        break;
      default:
        msg = args.join(' ');
    }

    if (msg) {
      console.warn(msg);
    }
  }
}

/**
 * Error
 *
 * @param { ...any } args
 */
function error(...args) {
  console.error(chalk.red.inverse(' error '), ...args);
}

/**
 * Fatal error
 *
 * @param { ...any } args
 */
function fatal(...args) {
  console.error(chalk.red.inverse(' fatal error '), ...args);
}

/**
 * Truncate 'string'
 *
 * @param { string } string
 * @returns { string }
 */
function truncate(string) {
  if (string.length > SEG_LENGTH * 1.5 + 3) {
    return string.slice(0, SEG_LENGTH) + '...' + string.slice(-SEG_LENGTH / 2);
  }

  return string;
}

var file = createCommonjsModule(function (module) {

/**
 * @typedef { import("http").ClientRequest } ClientRequest
 */

const {
  isAbsoluteFilePath,
  isCssRequest,
  isHtmlRequest,
  isJsRequest,
  isModule,
  isModuleBundlerFilePath,
  isNodeModuleFilePath
} = is;
const {
  warn,
  WARN_MISSING_EXTENSION,
  WARN_PACKAGE_INDEX,
  WARN_SERVER_TRANSPILE
} = log;
const { addHook } = pirates;



// Work around rollup-plugin-commonjs dynamic require
const loadModule = module$1._load;


const { URL } = url;

const FILE_TYPES = ['html', 'js', 'css'];
const IMPORT_EXTS = ['.js', '.mjs'];
const IMPORT_EXTS_TRANSPILER = ['.js', '.jsx', '.mjs', '.ts', '.tsx'];
const RE_GLOB = /[*[{]/;
const RE_TRANSPILER_HANDLES_SERVER = /\(\s?[a-zA-Z]+,\s?[a-zA-Z]+\s?\)/;
const RE_SEPARATOR = /[,;]\s?|\s/g;

let revertHook;

module.exports = {
  exists,
  expandPath,
  find,
  getAbsoluteProjectPath,
  getProjectPath,
  getTypeFromPath,
  getTypeFromRequest,
  importModule
};

/**
 * Validate that all file paths exist
 *
 * @param { string | Array<string> } filePaths
 * @returns { void }
 */
function exists(filePaths) {
  if (!Array.isArray(filePaths)) {
    filePaths = [filePaths];
  }

  for (const filePath of filePaths) {
    if (!fs.existsSync(path.resolve(filePath))) {
      throw Error(`path '${filePath}' does not exist`);
    }
  }
}

/**
 * Expand "filePath" into multiple filePaths
 * Handles globs and/or separators
 *
 * @param { string } filePath
 * @returns { Array<string> }
 */
function expandPath(filePath) {
  if (!filePath) {
    return;
  }

  if (typeof filePath === 'string' && fs.existsSync(path.resolve(filePath))) {
    return [filePath];
  }

  if (Array.isArray(filePath)) {
    return filePath.reduce((filePaths, fp) => {
      if (fp) {
        filePaths.push(...expandPath(fp));
      }
      return filePaths;
    }, []);
  }

  RE_SEPARATOR.lastIndex = 0;
  if (RE_SEPARATOR.test(filePath)) {
    filePath = filePath.split(RE_SEPARATOR);
  }
  if (!Array.isArray(filePath)) {
    filePath = [filePath];
  }

  return filePath.reduce((filePaths, fp) => {
    if (RE_GLOB.test(fp)) {
      filePaths.push(...glob.sync(fp));
    } else {
      filePaths.push(fp);
    }
    return filePaths;
  }, []);
}

/**
 * Find filePath for 'req'
 *
 * @param { ClientRequest | string } req
 * @param { object } [options]
 * @param { Array<string> } [options.directories]
 * @param { string } [options.type]
 * @returns { string | undefined }
 */
function find(req, { directories = config.directories, type } = {}) {
  const isRequestObject = typeof req !== 'string';
  const requestedFilePath = isRequestObject
    ? new URL(req.url, 'http://localhost').pathname
    : req;
  let filePath;

  if (!type) {
    type = isRequestObject ? getTypeFromRequest(req) : getTypeFromPath(req);
  }

  // Handle bundled js import
  if (isModuleBundlerFilePath(requestedFilePath)) {
    filePath = path.join(config.bundleDir, path.basename(requestedFilePath));
  } else if (isAbsoluteFilePath(requestedFilePath)) {
    filePath = resolveFilePath(requestedFilePath, type);
  } else {
    for (let directory of directories) {
      filePath = resolveFilePath(path.join(directory, requestedFilePath), type);

      if (filePath) {
        break;
      }
    }
  }

  if (!filePath) {
    return;
  }

  if (isRequestObject) {
    req.filePath = filePath;
    req.type = getTypeFromPath(filePath);
  }

  return filePath;
}

/**
 * Import esm/cjs module, transpiling if necessary (via require hook)
 *
 * @param { string } modulePath
 * @param { (filePath: string, isServer: boolean) => Promise<string> | string | undefined } [transpiler]
 * @returns { object }
 */
function importModule(modulePath, transpiler) {
  // Determine if transpiler supports transpiling server modules by checking number of arguments (filePath, isServer) handled
  const hasServerTranspiler =
    transpiler && RE_TRANSPILER_HANDLES_SERVER.test(transpiler.toString());

  if (revertHook !== undefined) {
    revertHook();
  }

  revertHook = addHook(
    (code, filePath) => {
      if (hasServerTranspiler) {
        const transpiled = transpiler(filePath, true);

        if (transpiled) {
          // Ignore async
          if (transpiled instanceof Promise) {
            warn(WARN_SERVER_TRANSPILE);
          } else {
            code = transpiled;
          }
        }
      }

      if (!isModule(code)) {
        return code;
      }

      return sucrase.transform(code, {
        transforms: ['imports'],
        filePath
      }).code;
    },
    {
      exts: transpiler ? IMPORT_EXTS_TRANSPILER : IMPORT_EXTS,
      ignoreNodeModules: false
    }
  );
  let mod = loadModule(modulePath, module, false);

  // Return default if only exported key
  if ('default' in mod && Object.keys(mod).length === 1) {
    mod = mod.default;
  }

  return mod;
}

/**
 * Retrieve the project relative path for "filePath"
 *
 * @param { string } filePath
 * @returns { string }
 */
function getProjectPath(filePath) {
  const projectFilePath = isAbsoluteFilePath(filePath)
    ? path.relative(process.cwd(), filePath)
    : filePath;

  return projectFilePath.startsWith('/')
    ? projectFilePath.slice(1)
    : projectFilePath;
}

/**
 * Retrieve the absolute path for the project relative path "filePath"
 *
 * @param { string } filePath
 * @returns { string }
 */
function getAbsoluteProjectPath(filePath) {
  return isAbsoluteFilePath(filePath)
    ? filePath
    : path.join(
        process.cwd(),
        filePath.charAt(0) === '/' ? filePath.slice(1) : filePath
      );
}

/**
 * Retrieve resource type
 *
 * @param { ClientRequest } req
 * @returns { string }
 */
function getTypeFromRequest(req) {
  if (req.type) {
    return req.type;
  } else if (isHtmlRequest(req)) {
    return 'html';
  } else if (isCssRequest(req)) {
    return 'css';
  } else if (isJsRequest(req)) {
    return 'js';
  } else {
    return '';
  }
}

/**
 * Retrieve generic file type from "filePath" extension
 *
 * @param { string } filePath
 * @returns { string }
 */
function getTypeFromPath(filePath) {
  return config.typesByExtension[path.extname(filePath)];
}

/**
 * Resolve "filePath" of "type"
 * Handles missing extensions and package indexes
 *
 * @param { string } filePath
 * @param { string } type
 * @returns { string }
 */
function resolveFilePath(filePath, type) {
  try {
    const stat = fs.statSync(filePath);

    if (stat.isFile()) {
      return filePath;
    }
  } catch (err) {
    // Not found, possibly no extension or package
  }

  if (!type) {
    for (type of FILE_TYPES) {
      const fp = resolveFilePath(filePath, type);

      if (fp) {
        return fp;
      }
    }
  }

  let fp = resolveFilePathExtension(filePath, config.extensionsByType[type]);

  if (fp) {
    if (!isNodeModuleFilePath(fp)) {
      warn(WARN_MISSING_EXTENSION, getProjectPath(filePath));
    }
    return fp;
  }

  fp = resolveFilePathExtension(
    path.join(filePath, 'index'),
    config.extensionsByType[type]
  );

  if (fp && type === 'js') {
    if (!isNodeModuleFilePath(fp)) {
      warn(WARN_PACKAGE_INDEX, getProjectPath(filePath));
    }
  }

  return fp;
}

/**
 * Resolve missing extension for "filePath"
 *
 * @param { string } filePath
 * @param { Array<string> } extensions
 * @returns { string }
 */
function resolveFilePathExtension(filePath, extensions) {
  for (const ext of extensions) {
    const fp = filePath + ext;

    if (fs.existsSync(fp)) {
      return fp;
    }
  }

  return '';
}
});
var file_1 = file.exists;
var file_2 = file.expandPath;
var file_3 = file.find;
var file_4 = file.getAbsoluteProjectPath;
var file_5 = file.getProjectPath;
var file_6 = file.getTypeFromPath;
var file_7 = file.getTypeFromRequest;
var file_8 = file.importModule;

/**
 * @typedef { object } Package
 * @property { object } aliases
 * @property { boolean } isNodeModule
 * @property { string } manifestPath
 * @property { string } main
 * @property { string } name
 * @property { string } path
 * @property { Array<string> } paths
 */

const { find } = file;
const { isRelativeFilePath: isRelativeFilePath$1 } = is;



const MAX_FILE_SYSTEM_DEPTH = 10;
const RE_TRAILING = /\/+$|\\+$/;

var _package = {
  getPackage,
  resolveAliasPath,
  resolvePackagePath
};

/**
 * Retrieve package details for "filePath"
 *
 * @param { string } filePath
 * @param { string } [packagePath]
 * @returns { Package | undefined }
 */
function getPackage(filePath, packagePath = resolvePackagePath(filePath)) {
  if (packagePath === undefined || !fs.existsSync(packagePath)) {
    return;
  }

  const manifestPath = path.join(packagePath, 'package.json');
  const pkg = {
    aliases: {},
    isNodeModule: packagePath !== process.cwd(),
    manifestPath,
    main: '',
    name: '',
    path: packagePath,
    paths: resolveNodeModules(packagePath),
    version: ''
  };

  try {
    const json = JSON.parse(fs.readFileSync(manifestPath, 'utf8'));
    const findOptions = {
      directories: [packagePath],
      type: 'js'
    };
    let main = find(json.module || json.main || 'index.js', findOptions);

    // Resolve "browser" aliasing
    if (json.browser) {
      if (typeof json.browser === 'string') {
        main = find(json.browser, findOptions);
      } else {
        for (let key in json.browser) {
          let value = json.browser[key];

          if (typeof value === 'string') {
            // If no extension, or not a relative path, it's a package reference
            if (path.extname(key) || isRelativeFilePath$1(key)) {
              key = find(key, findOptions);
            }
            if (path.extname(value) || isRelativeFilePath$1(value)) {
              value = find(value, findOptions);
            }

            if (key === main) {
              main = value;
            }
            pkg.aliases[key] = value;
          } else {
            // TODO: warn about disabled module?
          }
        }
      }
    }

    // Store "main" as alias
    if (main) {
      pkg.aliases[packagePath] = main;
      pkg.aliases[json.name] = main;
    }

    pkg.name = json.name;
    pkg.main = main;
    pkg.version = json.version;
  } catch (err) {
    // No package.json found
  }

  return pkg;
}

/**
 * Resolve alias for "filePath"
 *
 * @param { string } filePath
 * @param { Package } pkg
 * @returns { string }
 */
function resolveAliasPath(filePath, pkg) {
  // Follow chain of aliases
  // a => b; b => c; c => d
  while (filePath in pkg.aliases) {
    filePath = pkg.aliases[filePath];
  }

  return filePath;
}

/**
 * Retrieve path to nearest directory with package.json for "filePath"
 *
 * @param { string } filePath
 * @returns { string | undefined }
 */
function resolvePackagePath(filePath) {
  filePath = filePath.replace(RE_TRAILING, '');
  const cwd = process.cwd();
  const isNodeModule = filePath.includes('node_modules');
  let depth = MAX_FILE_SYSTEM_DEPTH;
  let dir = filePath;
  let parent = '';
  let root = cwd;

  // Set root to first node_modules dir if in node_modules
  if (isNodeModule) {
    const parts = filePath.split(path.sep);
    const idx = parts.indexOf('node_modules');

    root = parts.slice(0, idx + 1).join(path.sep);
  }

  while (true) {
    // Stop at directory with package.json
    if (fs.existsSync(path.join(dir, 'package.json'))) {
      return dir;
    }

    parent = path.dirname(dir);

    // Stop if we hit root or max file system depth
    // Convert to lowercase to fix problems on Windows
    if (
      dir === root ||
      !--depth ||
      parent.toLowerCase() === dir.toLowerCase()
    ) {
      break;
    }

    // Walk
    dir = parent;
  }
}

/**
 * Gather all node_modules directories reachable from "pkgPath"
 *
 * @param { string } pkgPath
 * @returns { Array<string> }
 */
function resolveNodeModules(pkgPath) {
  let dir = pkgPath;
  let dirs = [];
  let depth = MAX_FILE_SYSTEM_DEPTH;
  let parent;

  if (process.env.NODE_PATH !== undefined) {
    dirs = process.env.NODE_PATH.split(path.delimiter).map((dir) =>
      path.resolve(dir)
    );
  }

  while (true) {
    parent = path.dirname(dir);
    // Stop if we hit max file system depth or root
    // Convert to lowercase to fix problems on Windows
    if (!--depth || parent.toLowerCase() === dir.toLowerCase()) {
      break;
    }

    const nodeModulesPath = path.resolve(dir, 'node_modules');

    if (fs.existsSync(nodeModulesPath)) {
      dirs.push(nodeModulesPath);
    }

    // Walk
    dir = parent;
  }

  return dirs;
}

const {
  getPackage: getPackage$1,
  resolveAliasPath: resolveAliasPath$1,
  resolvePackagePath: resolvePackagePath$1
} = _package;
const { find: find$1, getProjectPath } = file;
const { isAbsoluteFilePath: isAbsoluteFilePath$1, isRelativeFilePath: isRelativeFilePath$2 } = is;


const cache = new Map();
const packageCache = new Map();

var resolver = {
  clearResolverCache,
  getCachedPackage,
  resolve
};

/**
 * Resolve file path for "id" relative to "fromFilePath",
 * where "id" can be an absolute path, relative path, or npm package id
 *
 * @param { string } id
 * @param { string } [fromFilePath]
 * @returns { string | undefined }
 */
function resolve(id, fromFilePath = path.resolve('index.js')) {
  if (!id) {
    return;
  }

  const key = getCacheKey(fromFilePath, id);

  if (cache.has(key)) {
    return cache.get(key);
  }

  const filePath = _resolve(path.dirname(fromFilePath), id);

  if (!filePath) {
    return;
  }

  cache.set(key, filePath);
  return filePath;
}

/**
 * Retrieve file path for "id" relative to "fromDirPath"
 *
 * @param { string } fromDirPath
 * @param { string } id
 * @returns { string | undefined }
 */
function _resolve(fromDirPath, id) {
  const pkg = getCachedPackage(fromDirPath);

  if (!pkg) {
    return;
  }

  let filePath = resolveAliasPath$1(
    isRelativeFilePath$2(id) ? path.join(fromDirPath, id) : id,
    pkg
  );

  if (isAbsoluteFilePath$1(filePath)) {
    filePath = resolveAliasPath$1(find$1(filePath, { type: 'js' }), pkg);

    // Allow package references to fall through
    if (!filePath || isAbsoluteFilePath$1(filePath)) {
      return filePath;
    }
  }

  // filePath must be a package reference, so restart search from each package dir working upwards
  id = filePath;

  for (const packageDirPath of pkg.paths) {
    if (fromDirPath !== packageDirPath) {
      filePath = path.join(packageDirPath, id);
      filePath = _resolve(filePath, filePath);

      if (filePath) {
        return filePath;
      }
    }
  }
}

/**
 * Retrieve cache key
 *
 * @param { string } fromPath
 * @param { string } id
 * @returns { string }
 */
function getCacheKey(fromPath, id) {
  return `${getProjectPath(fromPath)}:${id}`;
}

/**
 * Retrieve Package instance for "dir"
 *
 * @param { string } dir
 * @returns { Package }
 */
function getCachedPackage(dir) {
  const pkgPath = resolvePackagePath$1(dir);
  let pkg = packageCache.get(pkgPath);

  if (!pkg) {
    pkg = getPackage$1(dir, pkgPath);
    packageCache.set(pkgPath, pkg);
  }

  return pkg;
}

/**
 * Clear caches
 */
function clearResolverCache() {
  cache.clear();
  packageCache.clear();
}

const { resolve: resolve$1 } = resolver;

const { rollup } = rollup$1;

const ROLLUP_INPUT_DEFAULTS = {
  // Only bundle local package files
  external: (id, parent, isResolved) => {
    // Skip if already handled by plugin
    if (isResolved || parent.includes('?commonjs-proxy')) {
      return false;
    }
    return /^[^./\0]/.test(id);
  },
  plugins: [
    rollupPluginReplace({
      'process.env.NODE_ENV': `"${process.env.NODE_ENV}"` || '"development"'
    }),
    rollupPluginNodeResolve({
      mainFields: ['browser', 'module', 'main']
    }),
    rollupPluginJson(),
    rollupPluginCommonjs({
      sourceMap: false
    })
  ],
  treeshake: false
};
const ROLLUP_OUTPUT_DEFAULTS = {
  format: 'es',
  sourcemap: false
};

/**
 * Bundle package at 'id'
 *
 * @param { string } id
 * @param { string } outputpath
 * @param { object } overrideOptions
 * @param { (err: Error) => void } fn
 */
var bundleWorker = async function bundle(id, outputpath, overrideOptions, fn) {
  overrideOptions = parseOptions(overrideOptions);
  const input = resolve$1(id);

  if (!input) {
    return fn(Error(`unable to find file for bundling with id "${id}"`));
  }

  const inputOptions = {
    input,
    ...ROLLUP_INPUT_DEFAULTS,
    ...overrideOptions.input
  };
  const outputOptions = {
    file: outputpath,
    ...ROLLUP_OUTPUT_DEFAULTS,
    ...overrideOptions.output
  };

  try {
    const bundled = await rollup(inputOptions);
    await bundled.write(outputOptions);
    fn();
  } catch (err) {
    console.log(err);
    fn(err);
  }
};

/**
 * Parse override options
 *
 * @param { object } options
 * @returns { object }
 */
function parseOptions(options) {
  if (!options) {
    return { input: {}, output: {} };
  }

  const { input, treeshake, output = {}, watch, ...inputOverride } = options;
  const { file, format, sourcemap, ...outputOverride } = output;

  return { input: inputOverride, output: outputOverride };
}

module.exports = bundleWorker;
