'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var fs = _interopDefault(require('fs'));
var path = _interopDefault(require('path'));
var rimraf = _interopDefault(require('rimraf'));
var isModule$2 = _interopDefault(require('is-module'));
var util = _interopDefault(require('util'));
var chalk = _interopDefault(require('chalk'));
var pirates = _interopDefault(require('pirates'));
var glob = _interopDefault(require('glob'));
var module$1 = _interopDefault(require('module'));
var sucrase = _interopDefault(require('sucrase'));
var url$1 = _interopDefault(require('url'));
var rollupPluginCommonjs = _interopDefault(require('rollup-plugin-commonjs'));
var rollupPluginJson = _interopDefault(require('rollup-plugin-json'));
var rollupPluginReplace = _interopDefault(require('rollup-plugin-replace'));
var rollupPluginNodeResolve = _interopDefault(require('rollup-plugin-node-resolve'));
var rollup$1 = _interopDefault(require('rollup'));
var debug$a = _interopDefault(require('debug'));
var workerFarm = _interopDefault(require('worker-farm'));
var permessageDeflate = _interopDefault(require('permessage-deflate'));
var fayeWebsocket = _interopDefault(require('faye-websocket'));
var http = _interopDefault(require('http'));
var https = _interopDefault(require('https'));
var crypto = _interopDefault(require('crypto'));
var mime = _interopDefault(require('mime'));
var send = _interopDefault(require('send'));
var zlib = _interopDefault(require('zlib'));
var chokidar = _interopDefault(require('chokidar'));
var serverDestroy = _interopDefault(require('server-destroy'));
var getPort = _interopDefault(require('get-port'));

const TESTING = process.env.NODE_ENV === 'test';
// Prevent parallel test runs from reading from same cache
const DIR = `.dvlp${TESTING ? process.getuid() : ''}`;
const VERSION = process.env.DVLP_VERSION;

const dir = path.resolve(DIR);
const bundleDirName = `${path.join(DIR, `bundle-${VERSION}`)}`;
const bundleDir = path.resolve(bundleDirName);
const maxModuleBundlerWorkers = parseInt(process.env.BUNDLE_WORKERS, 10) || 0;
const port = process.env.PORT ? Number(process.env.PORT) : 8080;

// Work around rollup-plugin-commonjs require.main.filename
if (TESTING || process.env.DVLP_LAUNCHER === 'cmd') {
  const bundleDirExists = fs.existsSync(bundleDir);
  const dirExists = fs.existsSync(dir);
  const rm = dirExists && !bundleDirExists;

  if (rm) {
    const contents = fs.readdirSync(dir).map((item) => path.resolve(dir, item));

    for (const item of contents) {
      // Delete all subdirectories
      if (fs.statSync(item).isDirectory()) {
        rimraf.sync(item);
      }
    }
  }
  if (!dirExists) {
    fs.mkdirSync(dir);
  }
  if (!bundleDirExists) {
    fs.mkdirSync(bundleDir);
  }

  if (TESTING) {
    process.on('exit', () => {
      rimraf.sync(dir);
    });
  }
}

/**
 * @typedef { object } Config
 * @property { number } activePort,
 * @property { string } bundleDir,
 * @property { string } bundleDirName,
 * @property { Array<string> } directories,
 * @property { object } extensionsByType,
 * @property { Array<string> } extensionsByType.css,
 * @property { Array<string> } extensionsByType.html,
 * @property { Array<string> } extensionsByType.js,
 * @property { number } latency,
 * @property { string } maxAge,
 * @property { number } maxModuleBundlerWorkers,
 * @property { number } port,
 * @property { string } testing,
 * @property { object } typesByExtension,
 * @property { object } typesByExtension,
 */
var config = {
  activePort: port,
  bundleDir,
  bundleDirName,
  directories: [process.cwd()],
  extensionsByType: {
    css: ['.css', '.sass', '.scss', '.less', '.styl', '.stylus'],
    html: [
      '.html',
      '.htm',
      '.nunjs',
      '.nunjucks',
      '.hbs',
      '.handlebars',
      '.dust'
    ],
    js: ['.js', '.mjs', '.jsx', '.ts', '.tsx', '.json']
  },
  latency: 50,
  maxAge: '10m',
  maxModuleBundlerWorkers,
  port,
  rollupConfigPath: path.join(dir, 'rollup.config.js'),
  testing: TESTING,
  typesByExtension: {
    '.sass': 'css',
    '.scss': 'css',
    '.less': 'css',
    '.styl': 'css',
    '.stylus': 'css',
    '.html': 'html',
    '.htm': 'html',
    '.nunjs': 'html',
    '.nunjucks': 'html',
    '.hbs': 'html',
    '.handlebars': 'html',
    '.dust': 'html',
    '.mjs': 'js',
    '.json': 'js',
    '.jsx': 'js',
    '.ts': 'js',
    '.tsx': 'js'
  }
};

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

/**
 * @typedef { import("http").ClientRequest } ClientRequest
 */







const HAS_UTIL_IS_PROXY = util.types && util.types.isProxy;
const RE_BARE_IMPORT = /^[^./]/;
const RE_INVALID = /[<>:"|?*]/;
const RE_JSON = /.json$/i;
const RE_LOCALHOST = /localhost|127\.0\.0\.1/;
const RE_NODE_MODULES = /node_modules/;
const RE_TYPE_CSS = /text\/css/i;
const RE_TYPE_HTML = /text\/html/i;
const RE_TYPE_JS = /application\/javascript/i;

var is = {
  isAbsoluteFilePath,
  isBareImport,
  isCssFilePath,
  isCssRequest,
  isHtmlFilePath,
  isHtmlRequest,
  isInvalidFilePath,
  isJsFilePath,
  isJsRequest,
  isJsonFilePath,
  isLocalhost,
  isNodeModuleFilePath,
  isModule,
  isModuleBundlerFilePath,
  isProjectFilePath,
  isPromise,
  isProxy,
  isRelativeFilePath,
  isValidFilePath
};

/**
 * Determine if "filePath" is absolute
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isAbsoluteFilePath(filePath) {
  return (
    'string' == typeof filePath &&
    path.isAbsolute(filePath) &&
    // Only absolute if from root
    filePath.startsWith(process.cwd().slice(0, 5))
  );
}

/**
 * Determine if 'id' is referencing a node_module
 *
 * @param { string } id
 * @returns { boolean }
 */
function isBareImport(id) {
  return RE_BARE_IMPORT.test(id);
}

/**
 * Determine if 'filePath' is for a css file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isCssFilePath(filePath) {
  return config.extensionsByType.css.includes(path.extname(filePath));
}

/**
 * Determine if 'req' is for a css file
 *
 * @param { ClientRequest } req
 * @returns { boolean }
 */
function isCssRequest(req) {
  return (
    req.type === 'css' ||
    RE_TYPE_CSS.test(req.headers.accept) ||
    isCssFilePath(req.url)
  );
}

/**
 * Determine if 'filePath' is for an html file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isHtmlFilePath(filePath) {
  return config.extensionsByType.html.includes(path.extname(filePath));
}

/**
 * Determine if 'req' is for an html file
 *
 * @param { ClientRequest } req
 * @returns { boolean }
 */
function isHtmlRequest(req) {
  return (
    req.type === 'html' ||
    RE_TYPE_HTML.test(req.headers.accept) ||
    isHtmlFilePath(req.url)
  );
}

/**
 * Determine if 'filePath' is invalid
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isInvalidFilePath(filePath) {
  return RE_INVALID.test(filePath);
}

/**
 * Determine if 'filePath' is for a js file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isJsFilePath(filePath) {
  return config.extensionsByType.js.includes(path.extname(filePath));
}

/**
 * Determine if 'req' is for a js file
 *
 * @param { ClientRequest } req
 * @returns { boolean }
 */
function isJsRequest(req) {
  return (
    req.type === 'js' ||
    // Almost always '*/*'
    RE_TYPE_JS.test(req.headers.accept) ||
    isJsFilePath(req.url)
  );
}

/**
 * Determine if 'filePath' is for a js file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isJsonFilePath(filePath) {
  return RE_JSON.test(filePath);
}

/**
 * Determine if 'url' is localhost
 *
 * @param { string } url
 * @returns { boolean }
 */
function isLocalhost(url) {
  return RE_LOCALHOST.test(url);
}

/**
 * Determine if 'filePath' is in node_modules
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isNodeModuleFilePath(filePath) {
  return RE_NODE_MODULES.test(filePath);
}

/**
 * Determine if filePath or code is es module
 *
 * @param { string } filePathOrCode
 * @returns { boolean }
 */
function isModule(filePathOrCode) {
  if (isJsFilePath(filePathOrCode)) {
    filePathOrCode = fs.readFileSync(filePathOrCode, 'utf8');
  }
  return isModule$2(filePathOrCode);
}

/**
 * Determine if 'filePath' is for a bundled module file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isModuleBundlerFilePath(filePath) {
  return filePath.includes(config.bundleDirName);
}

/**
 * Determine if 'obj' is Proxy
 *
 * @param { object } obj
 * @returns { boolean }
 */
function isProxy(obj) {
  if (HAS_UTIL_IS_PROXY) {
    return util.types.isProxy(obj);
  } else {
    return obj instanceof Proxy;
  }
}

/**
 * Determine if 'obj' is Promise instance
 *
 * @param { object } obj
 * @returns { boolean }
 */
function isPromise(obj) {
  if (HAS_UTIL_IS_PROXY) {
    return util.types.isPromise(obj);
  } else {
    return obj instanceof Promise;
  }
}

/**
 * Determine if 'filePath' is in project
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isProjectFilePath(filePath) {
  if (!path.isAbsolute(filePath)) {
    filePath = path.resolve(filePath);
  }
  return filePath.includes(process.cwd()) && !isNodeModuleFilePath(filePath);
}

/**
 * Determine if "filePath" is relative
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isRelativeFilePath(filePath) {
  return 'string' == typeof filePath && filePath.startsWith('.');
}

/**
 * Determine if "filePath" is valid.
 * If relative, resolves against "fromFilePath".
 *
 * @param { string } filePath
 * @param { string } [fromDir]
 * @returns { boolean }
 */
function isValidFilePath(filePath, fromDir = process.cwd()) {
  if (isRelativeFilePath(filePath)) {
    filePath = path.join(fromDir, filePath);
  }

  try {
    const stat = fs.statSync(filePath);
    if (stat.isFile()) {
      return true;
    }
  } catch (err) {
    return false;
  }
}

const SEG_LENGTH = 80;
const WARN_BARE_IMPORT =
  '⚠️  re-writing bare imports because browsers do not support Node-style import identifiers';
const WARN_MISSING_EXTENSION =
  '⚠️  adding missing extensions because browsers can only resolve valid URL identifiers';
const WARN_NODE_PATH =
  '⚠️  re-writing NODE_PATH import because browsers can only resolve valid URL identifiers';
const WARN_PACKAGE_INDEX =
  '⚠️  adding missing package index.js because browsers do not support Node-style package identifiers';
const WARN_SERVER_TRANSPILE =
  '⚠️  ignoring async Promise returned when executing transpiler on a server file. Be sure to check the "isServer" argument before returning a transpiled value';

let maxBareImport = 1;
let maxMissingExtension = 1;
let maxNodePath = 1;
let maxPackageIndex = 1;
let maxServerTranspile = 1;
let silent = false;

var log = {
  WARN_BARE_IMPORT,
  WARN_MISSING_EXTENSION,
  WARN_NODE_PATH,
  WARN_PACKAGE_INDEX,
  WARN_SERVER_TRANSPILE,
  info,
  noisyInfo,
  warn,
  error,
  fatal,
  set silent(value) {
    silent = value;
  }
};

/**
 * Log if not testing/silent
 *
 * @param { ...any } args
 */
function info(...args) {
  if (!config.testing && !silent) {
    console.log(' ', ...args.map(truncate));
  }
}

/**
 * Log if not testing, even if silent
 *
 * @param { ...any } args
 */
function noisyInfo(...args) {
  if (!config.testing) {
    console.log(' ', ...args.map(truncate));
  }
}

/**
 * Warn if not testing/silent
 *
 * @param { ...any } args
 */
function warn(...args) {
  if (!config.testing) {
    const [warning] = args;
    let msg = '';

    switch (warning) {
      case WARN_BARE_IMPORT:
        if (maxBareImport-- > 0) {
          msg = `${WARN_BARE_IMPORT}`;
        }
        break;
      case WARN_NODE_PATH:
        if (maxNodePath-- > 0) {
          msg = `${WARN_NODE_PATH}`;
        }
        break;
      case WARN_MISSING_EXTENSION:
        if (maxMissingExtension-- > 0) {
          msg = `${WARN_MISSING_EXTENSION}`;
        }
        break;
      case WARN_PACKAGE_INDEX:
        if (maxPackageIndex-- > 0) {
          msg = `${WARN_PACKAGE_INDEX}`;
        }
        break;
      case WARN_SERVER_TRANSPILE:
        if (maxServerTranspile-- > 0) {
          msg = `${WARN_SERVER_TRANSPILE}`;
        }
        break;
      default:
        msg = args.join(' ');
    }

    if (msg) {
      console.warn(msg);
    }
  }
}

/**
 * Error
 *
 * @param { ...any } args
 */
function error(...args) {
  console.error(chalk.red.inverse(' error '), ...args);
}

/**
 * Fatal error
 *
 * @param { ...any } args
 */
function fatal(...args) {
  console.error(chalk.red.inverse(' fatal error '), ...args);
}

/**
 * Truncate 'string'
 *
 * @param { string } string
 * @returns { string }
 */
function truncate(string) {
  if (string.length > SEG_LENGTH * 1.5 + 3) {
    return string.slice(0, SEG_LENGTH) + '...' + string.slice(-SEG_LENGTH / 2);
  }

  return string;
}

var file = createCommonjsModule(function (module) {

/**
 * @typedef { import("http").ClientRequest } ClientRequest
 */

const {
  isAbsoluteFilePath,
  isCssRequest,
  isHtmlRequest,
  isJsRequest,
  isModule,
  isModuleBundlerFilePath,
  isNodeModuleFilePath
} = is;
const {
  warn,
  WARN_MISSING_EXTENSION,
  WARN_PACKAGE_INDEX,
  WARN_SERVER_TRANSPILE
} = log;
const { addHook } = pirates;



// Work around rollup-plugin-commonjs dynamic require
const loadModule = module$1._load;


const { URL } = url$1;

const FILE_TYPES = ['html', 'js', 'css'];
const IMPORT_EXTS = ['.js', '.mjs'];
const IMPORT_EXTS_TRANSPILER = ['.js', '.jsx', '.mjs', '.ts', '.tsx'];
const RE_GLOB = /[*[{]/;
const RE_TRANSPILER_HANDLES_SERVER = /\(\s?[a-zA-Z]+,\s?[a-zA-Z]+\s?\)/;
const RE_SEPARATOR = /[,;]\s?|\s/g;

let revertHook;

module.exports = {
  exists,
  expandPath,
  find,
  getAbsoluteProjectPath,
  getProjectPath,
  getTypeFromPath,
  getTypeFromRequest,
  importModule
};

/**
 * Validate that all file paths exist
 *
 * @param { string | Array<string> } filePaths
 * @returns { void }
 */
function exists(filePaths) {
  if (!Array.isArray(filePaths)) {
    filePaths = [filePaths];
  }

  for (const filePath of filePaths) {
    if (!fs.existsSync(path.resolve(filePath))) {
      throw Error(`path '${filePath}' does not exist`);
    }
  }
}

/**
 * Expand "filePath" into multiple filePaths
 * Handles globs and/or separators
 *
 * @param { string } filePath
 * @returns { Array<string> }
 */
function expandPath(filePath) {
  if (!filePath) {
    return;
  }

  if (typeof filePath === 'string' && fs.existsSync(path.resolve(filePath))) {
    return [filePath];
  }

  if (Array.isArray(filePath)) {
    return filePath.reduce((filePaths, fp) => {
      if (fp) {
        filePaths.push(...expandPath(fp));
      }
      return filePaths;
    }, []);
  }

  RE_SEPARATOR.lastIndex = 0;
  if (RE_SEPARATOR.test(filePath)) {
    filePath = filePath.split(RE_SEPARATOR);
  }
  if (!Array.isArray(filePath)) {
    filePath = [filePath];
  }

  return filePath.reduce((filePaths, fp) => {
    if (RE_GLOB.test(fp)) {
      filePaths.push(...glob.sync(fp));
    } else {
      filePaths.push(fp);
    }
    return filePaths;
  }, []);
}

/**
 * Find filePath for 'req'
 *
 * @param { ClientRequest | string } req
 * @param { object } [options]
 * @param { Array<string> } [options.directories]
 * @param { string } [options.type]
 * @returns { string | undefined }
 */
function find(req, { directories = config.directories, type } = {}) {
  const isRequestObject = typeof req !== 'string';
  const requestedFilePath = isRequestObject
    ? new URL(req.url, 'http://localhost').pathname
    : req;
  let filePath;

  if (!type) {
    type = isRequestObject ? getTypeFromRequest(req) : getTypeFromPath(req);
  }

  // Handle bundled js import
  if (isModuleBundlerFilePath(requestedFilePath)) {
    filePath = path.join(config.bundleDir, path.basename(requestedFilePath));
  } else if (isAbsoluteFilePath(requestedFilePath)) {
    filePath = resolveFilePath(requestedFilePath, type);
  } else {
    for (let directory of directories) {
      filePath = resolveFilePath(path.join(directory, requestedFilePath), type);

      if (filePath) {
        break;
      }
    }
  }

  if (!filePath) {
    return;
  }

  if (isRequestObject) {
    req.filePath = filePath;
    req.type = getTypeFromPath(filePath);
  }

  return filePath;
}

/**
 * Import esm/cjs module, transpiling if necessary (via require hook)
 *
 * @param { string } modulePath
 * @param { (filePath: string, isServer: boolean) => Promise<string> | string | undefined } [transpiler]
 * @returns { object }
 */
function importModule(modulePath, transpiler) {
  // Determine if transpiler supports transpiling server modules by checking number of arguments (filePath, isServer) handled
  const hasServerTranspiler =
    transpiler && RE_TRANSPILER_HANDLES_SERVER.test(transpiler.toString());

  if (revertHook !== undefined) {
    revertHook();
  }

  revertHook = addHook(
    (code, filePath) => {
      if (hasServerTranspiler) {
        const transpiled = transpiler(filePath, true);

        if (transpiled) {
          // Ignore async
          if (transpiled instanceof Promise) {
            warn(WARN_SERVER_TRANSPILE);
          } else {
            code = transpiled;
          }
        }
      }

      if (!isModule(code)) {
        return code;
      }

      return sucrase.transform(code, {
        transforms: ['imports'],
        filePath
      }).code;
    },
    {
      exts: transpiler ? IMPORT_EXTS_TRANSPILER : IMPORT_EXTS,
      ignoreNodeModules: false
    }
  );
  let mod = loadModule(modulePath, module, false);

  // Return default if only exported key
  if ('default' in mod && Object.keys(mod).length === 1) {
    mod = mod.default;
  }

  return mod;
}

/**
 * Retrieve the project relative path for "filePath"
 *
 * @param { string } filePath
 * @returns { string }
 */
function getProjectPath(filePath) {
  const projectFilePath = isAbsoluteFilePath(filePath)
    ? path.relative(process.cwd(), filePath)
    : filePath;

  return projectFilePath.startsWith('/')
    ? projectFilePath.slice(1)
    : projectFilePath;
}

/**
 * Retrieve the absolute path for the project relative path "filePath"
 *
 * @param { string } filePath
 * @returns { string }
 */
function getAbsoluteProjectPath(filePath) {
  return isAbsoluteFilePath(filePath)
    ? filePath
    : path.join(
        process.cwd(),
        filePath.charAt(0) === '/' ? filePath.slice(1) : filePath
      );
}

/**
 * Retrieve resource type
 *
 * @param { ClientRequest } req
 * @returns { string }
 */
function getTypeFromRequest(req) {
  if (req.type) {
    return req.type;
  } else if (isHtmlRequest(req)) {
    return 'html';
  } else if (isCssRequest(req)) {
    return 'css';
  } else if (isJsRequest(req)) {
    return 'js';
  } else {
    return '';
  }
}

/**
 * Retrieve generic file type from "filePath" extension
 *
 * @param { string } filePath
 * @returns { string }
 */
function getTypeFromPath(filePath) {
  return config.typesByExtension[path.extname(filePath)];
}

/**
 * Resolve "filePath" of "type"
 * Handles missing extensions and package indexes
 *
 * @param { string } filePath
 * @param { string } type
 * @returns { string }
 */
function resolveFilePath(filePath, type) {
  try {
    const stat = fs.statSync(filePath);

    if (stat.isFile()) {
      return filePath;
    }
  } catch (err) {
    // Not found, possibly no extension or package
  }

  if (!type) {
    for (type of FILE_TYPES) {
      const fp = resolveFilePath(filePath, type);

      if (fp) {
        return fp;
      }
    }
  }

  let fp = resolveFilePathExtension(filePath, config.extensionsByType[type]);

  if (fp) {
    if (!isNodeModuleFilePath(fp)) {
      warn(WARN_MISSING_EXTENSION, getProjectPath(filePath));
    }
    return fp;
  }

  fp = resolveFilePathExtension(
    path.join(filePath, 'index'),
    config.extensionsByType[type]
  );

  if (fp && type === 'js') {
    if (!isNodeModuleFilePath(fp)) {
      warn(WARN_PACKAGE_INDEX, getProjectPath(filePath));
    }
  }

  return fp;
}

/**
 * Resolve missing extension for "filePath"
 *
 * @param { string } filePath
 * @param { Array<string> } extensions
 * @returns { string }
 */
function resolveFilePathExtension(filePath, extensions) {
  for (const ext of extensions) {
    const fp = filePath + ext;

    if (fs.existsSync(fp)) {
      return fp;
    }
  }

  return '';
}
});
var file_1 = file.exists;
var file_2 = file.expandPath;
var file_3 = file.find;
var file_4 = file.getAbsoluteProjectPath;
var file_5 = file.getProjectPath;
var file_6 = file.getTypeFromPath;
var file_7 = file.getTypeFromRequest;
var file_8 = file.importModule;

/**
 * @typedef { object } Package
 * @property { object } aliases
 * @property { boolean } isNodeModule
 * @property { string } manifestPath
 * @property { string } main
 * @property { string } name
 * @property { string } path
 * @property { Array<string> } paths
 */

const { find } = file;
const { isRelativeFilePath: isRelativeFilePath$1 } = is;



const MAX_FILE_SYSTEM_DEPTH = 10;
const RE_TRAILING = /\/+$|\\+$/;

var _package = {
  getPackage,
  resolveAliasPath,
  resolvePackagePath
};

/**
 * Retrieve package details for "filePath"
 *
 * @param { string } filePath
 * @param { string } [packagePath]
 * @returns { Package | undefined }
 */
function getPackage(filePath, packagePath = resolvePackagePath(filePath)) {
  if (packagePath === undefined || !fs.existsSync(packagePath)) {
    return;
  }

  const manifestPath = path.join(packagePath, 'package.json');
  const pkg = {
    aliases: {},
    isNodeModule: packagePath !== process.cwd(),
    manifestPath,
    main: '',
    name: '',
    path: packagePath,
    paths: resolveNodeModules(packagePath),
    version: ''
  };

  try {
    const json = JSON.parse(fs.readFileSync(manifestPath, 'utf8'));
    const findOptions = {
      directories: [packagePath],
      type: 'js'
    };
    let main = find(json.module || json.main || 'index.js', findOptions);

    // Resolve "browser" aliasing
    if (json.browser) {
      if (typeof json.browser === 'string') {
        main = find(json.browser, findOptions);
      } else {
        for (let key in json.browser) {
          let value = json.browser[key];

          if (typeof value === 'string') {
            // If no extension, or not a relative path, it's a package reference
            if (path.extname(key) || isRelativeFilePath$1(key)) {
              key = find(key, findOptions);
            }
            if (path.extname(value) || isRelativeFilePath$1(value)) {
              value = find(value, findOptions);
            }

            if (key === main) {
              main = value;
            }
            pkg.aliases[key] = value;
          } else {
            // TODO: warn about disabled module?
          }
        }
      }
    }

    // Store "main" as alias
    if (main) {
      pkg.aliases[packagePath] = main;
      pkg.aliases[json.name] = main;
    }

    pkg.name = json.name;
    pkg.main = main;
    pkg.version = json.version;
  } catch (err) {
    // No package.json found
  }

  return pkg;
}

/**
 * Resolve alias for "filePath"
 *
 * @param { string } filePath
 * @param { Package } pkg
 * @returns { string }
 */
function resolveAliasPath(filePath, pkg) {
  // Follow chain of aliases
  // a => b; b => c; c => d
  while (filePath in pkg.aliases) {
    filePath = pkg.aliases[filePath];
  }

  return filePath;
}

/**
 * Retrieve path to nearest directory with package.json for "filePath"
 *
 * @param { string } filePath
 * @returns { string | undefined }
 */
function resolvePackagePath(filePath) {
  filePath = filePath.replace(RE_TRAILING, '');
  const cwd = process.cwd();
  const isNodeModule = filePath.includes('node_modules');
  let depth = MAX_FILE_SYSTEM_DEPTH;
  let dir = filePath;
  let parent = '';
  let root = cwd;

  // Set root to first node_modules dir if in node_modules
  if (isNodeModule) {
    const parts = filePath.split(path.sep);
    const idx = parts.indexOf('node_modules');

    root = parts.slice(0, idx + 1).join(path.sep);
  }

  while (true) {
    // Stop at directory with package.json
    if (fs.existsSync(path.join(dir, 'package.json'))) {
      return dir;
    }

    parent = path.dirname(dir);

    // Stop if we hit root or max file system depth
    // Convert to lowercase to fix problems on Windows
    if (
      dir === root ||
      !--depth ||
      parent.toLowerCase() === dir.toLowerCase()
    ) {
      break;
    }

    // Walk
    dir = parent;
  }
}

/**
 * Gather all node_modules directories reachable from "pkgPath"
 *
 * @param { string } pkgPath
 * @returns { Array<string> }
 */
function resolveNodeModules(pkgPath) {
  let dir = pkgPath;
  let dirs = [];
  let depth = MAX_FILE_SYSTEM_DEPTH;
  let parent;

  if (process.env.NODE_PATH !== undefined) {
    dirs = process.env.NODE_PATH.split(path.delimiter).map((dir) =>
      path.resolve(dir)
    );
  }

  while (true) {
    parent = path.dirname(dir);
    // Stop if we hit max file system depth or root
    // Convert to lowercase to fix problems on Windows
    if (!--depth || parent.toLowerCase() === dir.toLowerCase()) {
      break;
    }

    const nodeModulesPath = path.resolve(dir, 'node_modules');

    if (fs.existsSync(nodeModulesPath)) {
      dirs.push(nodeModulesPath);
    }

    // Walk
    dir = parent;
  }

  return dirs;
}

const {
  getPackage: getPackage$1,
  resolveAliasPath: resolveAliasPath$1,
  resolvePackagePath: resolvePackagePath$1
} = _package;
const { find: find$1, getProjectPath } = file;
const { isAbsoluteFilePath: isAbsoluteFilePath$1, isRelativeFilePath: isRelativeFilePath$2 } = is;


const cache = new Map();
const packageCache = new Map();

var resolver = {
  clearResolverCache,
  getCachedPackage,
  resolve
};

/**
 * Resolve file path for "id" relative to "fromFilePath",
 * where "id" can be an absolute path, relative path, or npm package id
 *
 * @param { string } id
 * @param { string } [fromFilePath]
 * @returns { string | undefined }
 */
function resolve(id, fromFilePath = path.resolve('index.js')) {
  if (!id) {
    return;
  }

  const key = getCacheKey(fromFilePath, id);

  if (cache.has(key)) {
    return cache.get(key);
  }

  const filePath = _resolve(path.dirname(fromFilePath), id);

  if (!filePath) {
    return;
  }

  cache.set(key, filePath);
  return filePath;
}

/**
 * Retrieve file path for "id" relative to "fromDirPath"
 *
 * @param { string } fromDirPath
 * @param { string } id
 * @returns { string | undefined }
 */
function _resolve(fromDirPath, id) {
  const pkg = getCachedPackage(fromDirPath);

  if (!pkg) {
    return;
  }

  let filePath = resolveAliasPath$1(
    isRelativeFilePath$2(id) ? path.join(fromDirPath, id) : id,
    pkg
  );

  if (isAbsoluteFilePath$1(filePath)) {
    filePath = resolveAliasPath$1(find$1(filePath, { type: 'js' }), pkg);

    // Allow package references to fall through
    if (!filePath || isAbsoluteFilePath$1(filePath)) {
      return filePath;
    }
  }

  // filePath must be a package reference, so restart search from each package dir working upwards
  id = filePath;

  for (const packageDirPath of pkg.paths) {
    if (fromDirPath !== packageDirPath) {
      filePath = path.join(packageDirPath, id);
      filePath = _resolve(filePath, filePath);

      if (filePath) {
        return filePath;
      }
    }
  }
}

/**
 * Retrieve cache key
 *
 * @param { string } fromPath
 * @param { string } id
 * @returns { string }
 */
function getCacheKey(fromPath, id) {
  return `${getProjectPath(fromPath)}:${id}`;
}

/**
 * Retrieve Package instance for "dir"
 *
 * @param { string } dir
 * @returns { Package }
 */
function getCachedPackage(dir) {
  const pkgPath = resolvePackagePath$1(dir);
  let pkg = packageCache.get(pkgPath);

  if (!pkg) {
    pkg = getPackage$1(dir, pkgPath);
    packageCache.set(pkgPath, pkg);
  }

  return pkg;
}

/**
 * Clear caches
 */
function clearResolverCache() {
  cache.clear();
  packageCache.clear();
}

const { resolve: resolve$1 } = resolver;

const { rollup } = rollup$1;

const ROLLUP_INPUT_DEFAULTS = {
  // Only bundle local package files
  external: (id, parent, isResolved) => {
    // Skip if already handled by plugin
    if (isResolved || parent.includes('?commonjs-proxy')) {
      return false;
    }
    return /^[^./\0]/.test(id);
  },
  plugins: [
    rollupPluginReplace({
      'process.env.NODE_ENV': `"${process.env.NODE_ENV}"` || '"development"'
    }),
    rollupPluginNodeResolve({
      mainFields: ['browser', 'module', 'main']
    }),
    rollupPluginJson(),
    rollupPluginCommonjs({
      sourceMap: false
    })
  ],
  treeshake: false
};
const ROLLUP_OUTPUT_DEFAULTS = {
  format: 'es',
  sourcemap: false
};

/**
 * Bundle package at 'id'
 *
 * @param { string } id
 * @param { string } outputpath
 * @param { object } overrideOptions
 * @param { (err: Error) => void } fn
 */
var bundleWorker = async function bundle(id, outputpath, overrideOptions, fn) {
  overrideOptions = parseOptions(overrideOptions);
  const input = resolve$1(id);

  if (!input) {
    return fn(Error(`unable to find file for bundling with id "${id}"`));
  }

  const inputOptions = {
    input,
    ...ROLLUP_INPUT_DEFAULTS,
    ...overrideOptions.input
  };
  const outputOptions = {
    file: outputpath,
    ...ROLLUP_OUTPUT_DEFAULTS,
    ...overrideOptions.output
  };

  try {
    const bundled = await rollup(inputOptions);
    await bundled.write(outputOptions);
    fn();
  } catch (err) {
    console.log(err);
    fn(err);
  }
};

/**
 * Parse override options
 *
 * @param { object } options
 * @returns { object }
 */
function parseOptions(options) {
  if (!options) {
    return { input: {}, output: {} };
  }

  const { input, treeshake, output = {}, watch, ...inputOverride } = options;
  const { file, format, sourcemap, ...outputOverride } = output;

  return { input: inputOverride, output: outputOverride };
}

const timers = {};

var stopwatch = {
  start,
  pause,
  stop,
  clear,
  msDiff
};

/**
 * Start timer with 'id'
 *
 * @param { string } id
 */
function start(id) {
  if (!timers[id]) {
    timers[id] = {
      start: 0,
      elapsed: 0
    };
  }
  timers[id].start = process.hrtime();
}

/**
 * Pause timer with 'id'
 *
 * @param { string } id
 */
function pause(id) {
  if (!timers[id]) {
    return start(id);
  }

  timers[id].elapsed += msDiff(process.hrtime(), timers[id].start);
}

/**
 * Stop timer with 'id' and return elapsed
 *
 * @param { string } id
 * @param { boolean } formatted
 * @param { boolean } padded
 * @returns { number | string }
 */
function stop(id, formatted, padded) {
  if (!timers[id]) {
    return formatted ? format(0, padded ? 5 : 0) : 0;
  }

  const elapsed =
    timers[id].elapsed + msDiff(process.hrtime(), timers[id].start);

  clear(id);
  return formatted ? format(elapsed, padded ? 5 : 0) : elapsed;
}

/**
 * clear timer with 'id'
 *
 * @param { string } id
 */
function clear(id) {
  delete timers[id];
}

/**
 * Retrieve difference in ms
 *
 * @param { Array<number> } t1
 * @param { Array<number> } t2
 * @returns { number }
 */
function msDiff(t1, t2) {
  t1 = (t1[0] * 1e9 + t1[1]) / 1e6;
  t2 = (t2[0] * 1e9 + t2[1]) / 1e6;
  return Math.ceil((t1 - t2) * 100) / 100;
}

/**
 * Format 'duration'
 *
 * @param { number } duration
 * @param { number } [length]
 * @returns { string }
 */
function format(duration, length) {
  const colour = duration > 10 ? (duration > 100 ? 'red' : 'yellow') : 'green';
  let formatted =
    duration < 1000
      ? `${Math.round(duration)}ms`
      : `${Math.floor((duration / 1000) * 100) / 100}s`;

  if (length && 'padStart' in String.prototype) {
    formatted = formatted.padStart(length, ' ');
  }

  return chalk[colour](formatted);
}

const { info: info$1, error: error$1 } = log;
const {
  isJsFilePath: isJsFilePath$1,
  isNodeModuleFilePath: isNodeModuleFilePath$1,
  isPromise: isPromise$1
} = is;



const debug = debug$a('dvlp:module');

const { getCachedPackage: getCachedPackage$1 } = resolver;




const cache$1 = new Map();
let workers;

if (config.maxModuleBundlerWorkers) {
  debug(`bundling modules with ${config.maxModuleBundlerWorkers} workers`);
}

if (fs.existsSync(config.bundleDir)) {
  const names = new Map();

  fs.readdirSync(config.bundleDir)
    .filter(isJsFilePath$1)
    .forEach((resolvedId) => {
      const name = resolvedId.slice(0, resolvedId.lastIndexOf('-'));

      if (!names.has(name)) {
        cache$1.set(resolvedId, true);
        names.set(name, resolvedId);
      } else {
        // Clear instances if duplicates with different versions
        const existing = names.get(name);

        cache$1.delete(existing);
        fs.unlinkSync(path.resolve(config.bundleDir, existing));
        fs.unlinkSync(path.resolve(config.bundleDir, resolvedId));
      }
    });
}

var bundler_1 = {
  bundle,
  cleanBundles,
  destroyWorkers,
  resolveModuleId,
  resolveModulePath
};

/**
 * Resolve module id into cacheable id
 *
 * @param { string } id
 * @param { string } filePath
 * @returns { string }
 */
function resolveModuleId(id, filePath) {
  if (!isNodeModuleFilePath$1(filePath)) {
    return '';
  }

  const pkg = getCachedPackage$1(path.dirname(filePath));

  return `${encodeId(id)}-${pkg.version}.js`;
}

/**
 * Retrieve path to cached 'resolvedId'
 *
 * @param { string } resolvedId
 * @returns { string }
 */
function resolveModulePath(resolvedId) {
  return path.join(config.bundleDir, resolvedId);
}

/**
 * Trigger bundle of 'id'
 *
 * @param { string } resolvedId
 * @param { string } [id]
 * @param { object } [rollupConfig]
 * @returns { Promise<string> }
 */
function bundle(resolvedId, id, rollupConfig) {
  if (!resolvedId) {
    return null;
  }
  if (!id) {
    id = decodeId(resolvedId.slice(0, resolvedId.lastIndexOf('-')));
  }
  const filePath = resolveModulePath(resolvedId);

  if (!cache$1.has(resolvedId)) {
    return doBundle(id, resolvedId, filePath, rollupConfig);
  } else if (isPromise$1(cache$1.get(resolvedId))) {
    return cache$1.get(resolvedId);
  } else {
    return Promise.resolve(filePath);
  }
}

/**
 * Bundle module at 'id'
 *
 * @param { string } id
 * @param { string } resolvedId
 * @param { string } filePath
 * @param { object } [rollupConfig]
 * @returns { Promise<string> }
 */
function doBundle(id, resolvedId, filePath, rollupConfig) {
  const promiseToCache = new Promise(async (resolve, reject) => {
    stopwatch.start(id);

    getBundler()(id, filePath, rollupConfig, (err) => {
      if (err) {
        error$1(`unable to bundle ${id}`);
        cache$1.delete(resolvedId);
        return reject(err);
      }

      // Can't use file.getProjectPath() here because of circular dependency
      info$1(
        `${stopwatch.stop(id, true, true)} bundled ${chalk.green(
          id
        )} as ${chalk.green(path.relative(process.cwd(), filePath))}`
      );
      cache$1.set(resolvedId, true);
      resolve(filePath);
    });
  });

  cache$1.set(resolvedId, promiseToCache);
  return promiseToCache;
}

/**
 * Retrieve bundler.
 * Starts workers if config.maxModuleBundlerWorkers > 0,
 * otherwise uses bundler directly
 *
 * @returns { (id: string, outputPath: string, overrideOptions: object, fn: (err: Error) => void) => void }
 */
function getBundler() {
  if (!config.maxModuleBundlerWorkers) {
    return bundleWorker;
  }

  if (!workers) {
    workers = workerFarm(
      { maxConcurrentWorkers: config.maxModuleBundlerWorkers },
      path.resolve(__dirname, './bundle-worker.js')
    );
    debug(`spawned ${config.maxModuleBundlerWorkers} bundler workers`);
  }

  return workers;
}

/**
 * Clear memory and disk cache
 */
function cleanBundles() {
  for (const resolvedId of cache$1) {
    try {
      fs.unlinkSync(resolveModulePath(resolvedId));
    } catch (err) {
      // ignore
    } finally {
      cache$1.delete(resolvedId);
    }
  }
}

/**
 * Terminate workers
 */
function destroyWorkers() {
  return new Promise((resolve, reject) => {
    if (!workers) {
      return resolve();
    }

    workerFarm.end(workers, (msg) => {
      workers = undefined;
      if (msg) {
        return reject(Error(msg));
      }
      resolve();
    });
  });
}

function encodeId(id) {
  return id.replace(/\//g, '__');
}

function decodeId(id) {
  return id.replace(/__/g, '/');
}

const { URL } = url$1;

const RE_WEB_SOCKET = /wss?:/;

var url = {
  filePathToUrl,
  getUrl,
  getUrlCacheKey,
  isWebSocketUrl
};

/**
 * Determine if 'url' is a WebSocket
 *
 * @param { URL } url
 * @returns { boolean }
 */
function isWebSocketUrl(url) {
  return RE_WEB_SOCKET.test(url.protocol);
}

/**
 * Retrieve URL instance from 'req'
 *
 * @param { string | MockRequest | ClientRequest | URL } req
 * @returns { URL }
 */
function getUrl(req) {
  if (!(req instanceof URL)) {
    req = new URL(
      typeof req === 'string' ? decodeURIComponent(req) : req.url,
      `http://localhost:${config.activePort}`
    );
  }
  // Map loopback address to localhost
  if (req.hostname === '127.0.0.1') {
    req.hostname = 'localhost';
  }

  return req;
}

/**
 * Retrieve key for 'url'
 *
 * @param { URL } url
 * @returns { string }
 * @private
 */
function getUrlCacheKey(url) {
  // Map loopback address to localhost
  const host = url.host === '127.0.0.1' ? 'localhost' : url.host;
  let key = `${host}${url.pathname}`;

  if (key.endsWith('/')) {
    key = key.slice(0, -1);
  }

  return key;
}

/**
 * Convert file path to valid url
 * Handles platform differences
 *
 * @param { string } filePath
 * @returns { string }
 */
function filePathToUrl(filePath) {
  return filePath.replace(/\\/g, '/');
}

/**
 * @typedef { object } PushStream
 * @property { string } url
 * @property { string } type
 */
/**
 * @typedef { object } PushEvent
 * @property { string | object } message
 * @property { object } [options]
 * @property { string } [options.event]
 * @property { string } [options.id]
 */

const { isWebSocketUrl: isWebSocketUrl$1, getUrl: getUrl$1, getUrlCacheKey: getUrlCacheKey$1 } = url;
const debug$1 = debug$a('dvlp:push');

const { error: error$2 } = log;

const { EventSource } = fayeWebsocket;

const DEFAULT_ES_CONFIG = {
  headers: { 'Access-Control-Allow-Origin': '*' },
  ping: 15,
  retry: 10
};
const RE_SOCKETIO_PROTOCOL = /socket\.?io|EIO/;

/**
 * @type { Map<string, Set<object>> }
 */
const cache$2 = new Map();

var pushEvents = {
  connectClient,
  destroyClients,
  pushEvent
};

/**
 * Initialize EventSource/WebSocket client
 *
 * @param { string | PushStream } stream
 * @param { ...any } args
 * @returns { void }
 */
function connectClient(stream, ...args) {
  const { type, url } = getStream(stream);
  const cacheKey = getUrlCacheKey$1(getUrl$1(url));
  const clients = cache$2.get(cacheKey) || new Set();
  let client;

  if (type === 'ws') {
    const [req, socket, body] = args;
    const isSocketio = RE_SOCKETIO_PROTOCOL.test(req.url);
    const extensionsHeaders = req.headers['Sec-WebSocket-Extensions'];
    const extensions =
      extensionsHeaders && extensionsHeaders.includes('permessage-deflate')
        ? [permessageDeflate]
        : [];

    client = new fayeWebsocket(req, socket, body, [], { extensions });
    client.on('message', (event) => {
      debug$1('received ws message', event.data);

      // Handle Socket.io channel protocol
      // ex: 40/channel?somequery=foo
      if (isSocketio) {
        const [packetAndChannel] = event.data.split('?');

        // Send separate packet and packet/channel responses
        if (packetAndChannel.includes('/')) {
          client.send(packetAndChannel.slice(0, packetAndChannel.indexOf('/')));
          client.send(packetAndChannel);
        }
      }
    });
    if (isSocketio) {
      client.send(
        '0{"sid":"dvlp","upgrades":[],"pingInterval":250000,"pingTimeout":600000}'
      );
    }
  } else {
    const [req, res] = args;

    client = new EventSource(req, res, DEFAULT_ES_CONFIG);
  }

  clients.add(client);
  cache$2.set(cacheKey, clients);
  debug$1(`added ${type} connection`, clients.size);

  client.on('close', () => {
    clients.delete(client);
    if (!clients.size) {
      cache$2.delete(cacheKey);
    }
    debug$1(`removed ${type} connection`, cache$2.size);
  });
}

/**
 * Push event data to WebSocket/EventSource clients
 *
 * @param { string | PushStream } stream
 * @param { PushEvent } event
 * @returns { void }
 */
function pushEvent(stream, event) {
  if (!stream || !event) {
    return;
  }

  const { url, type } = getStream(stream);
  const cacheKey = getUrlCacheKey$1(getUrl$1(url));
  const clients = cache$2.get(cacheKey);

  if (clients === undefined) {
    return error$2(`no push clients registered for ${url}`);
  }

  let { message, options } = event;

  if (!(typeof message === 'string')) {
    try {
      message = JSON.stringify(message);
    } catch (err) {
      return error$2(`unable to stringify message for push event`, message);
    }
  }
  if (type === 'ws' && options !== undefined) {
    const { event = '', namespace = '/', protocol } = options;

    // Handle socket.io protocol
    // https://github.com/socketio/socket.io-protocol/blob/master/Readme.md
    if (protocol && RE_SOCKETIO_PROTOCOL.test(protocol)) {
      message = `42${namespace},["${event}",${message}]`;
      // message = `${Buffer.from(message).length}:${message}`;
    }
    options = undefined;
  }

  debug$1(
    `push to ${clients.size} client${
      clients.size > 1 ? 's' : ''
    } connected on ${url}`
  );
  debug$1(message);
  for (const client of clients) {
    client.send(message, options);
  }
}

/**
 * Destroy all active push clients for connection at 'url'
 * If 'url' not defined, destroys all clients for all connections
 *
 * @param { string | PushStream } [stream]
 * @returns { void }
 */
function destroyClients(stream) {
  if (stream === undefined) {
    for (const cacheKey of cache$2.keys()) {
      destroyClient(cacheKey);
    }
    return;
  }

  const { url } = getStream(stream);
  const cacheKey = getUrlCacheKey$1(getUrl$1(url));

  destroyClient(cacheKey);
}

/**
 * Destroy client at 'cacheKey'
 *
 * @param { string } cacheKey
 * @returns { void }
 * @private
 */
function destroyClient(cacheKey) {
  const clients = cache$2.get(cacheKey);

  if (clients !== undefined) {
    for (const client of clients) {
      client.removeAllListeners();
      client.close();
    }
    clients.clear();
    cache$2.delete(cacheKey);
  }
}

/**
 * Retrieve PushStream from 'stream'
 * If passed as string, will determine type from url
 *
 * @param { string | PushStream } stream
 * @returns { PushStream }
 * @private
 */
function getStream(stream) {
  if (typeof stream === 'string') {
    const url = getUrl$1(stream);
    const isWebSocket = isWebSocketUrl$1(url);

    stream = {
      url: url.href,
      type: isWebSocket ? 'ws' : 'es'
    };
  }

  return stream;
}

/**
 * @typedef { import("http").ClientRequest } ClientRequest
 * @typedef { import("http").ServerResponse } ServerResponse
 * @typedef { import("../mock/index.js") } Mock
 */



const { connectClient: connectClient$1, pushEvent: pushEvent$1 } = pushEvents;
const { noisyInfo: noisyInfo$1 } = log;

const { URL: URL$1 } = url$1;


const { EventSource: EventSource$1 } = fayeWebsocket;

var common = {
  handleMockResponse,
  handleMockWebSocket,
  handlePushEvent
};

/**
 * Handle mock responses, including EventSource connection
 * Returns 'true' if handled
 *
 * @param { ClientRequest } req
 * @param { ServerResponse } res
 * @param { Mock } mocks
 * @returns { boolean }
 */
function handleMockResponse(req, res, mocks) {
  const url = new URL$1(req.url, `http://localhost:${config.activePort}`);
  let mock = url.searchParams.get('dvlpmock');

  if (mocks && mock) {
    mock = decodeURIComponent(mock);

    if (EventSource$1.isEventSource(req)) {
      connectClient$1(
        {
          url: mock,
          type: 'es'
        },
        req,
        res
      );
      // Send 'connect' event if it exists
      mocks.matchPushEvent(mock, 'connect', pushEvent$1);
      noisyInfo$1(
        `${stopwatch.stop(
          req.url,
          true,
          true
        )} connected to EventSource client at ${chalk.green(mock)}`
      );
    } else {
      mocks.matchResponse(mock, res);
    }

    return true;
  }

  return false;
}

/**
 * Handle mock WebSocket connection
 *
 * @param { ClientRequest } req
 * @param { object } socket
 * @param { object } body
 * @param { Mock } mocks
 * @returns { void }
 */
function handleMockWebSocket(req, socket, body, mocks) {
  const url = new URL$1(req.url, `http://localhost:${config.activePort}`);
  let mock = url.searchParams.get('dvlpmock');

  if (mocks && mock && fayeWebsocket.isWebSocket(req)) {
    mock = decodeURIComponent(mock);
    connectClient$1(
      {
        url: mock,
        type: 'ws'
      },
      req,
      socket,
      body
    );
    // Send 'connect' event if it exists
    mocks.matchPushEvent(mock, 'connect', pushEvent$1);
    noisyInfo$1(
      `${stopwatch.stop(
        req.url,
        true,
        true
      )} connected to WebSocket client at ${chalk.green(mock)}`
    );
  }
}

/**
 * Handle push event request
 * Returns 'true' if handled
 *
 * @param { ClientRequst } req
 * @param { ServerResponse } res
 * @param { Mock } mocks
 * @returns { boolean }
 */
function handlePushEvent(req, res, mocks) {
  if (req.method === 'POST' && req.url === '/dvlp/push-event') {
    let body = '';

    req.on('data', (chunk) => {
      body += chunk.toString();
    });
    req.on('end', () => {
      const { stream, event } = JSON.parse(body);

      if (typeof event === 'string') {
        mocks.matchPushEvent(stream, event, pushEvent$1);
      } else {
        pushEvent$1(stream, event);
      }

      res.writeHead(200);
      res.end('ok');
    });

    return true;
  }

  return false;
}

/**
 * @typedef { import("http").ClientRequest } ClientRequest
 */




const { isLocalhost: isLocalhost$1, isProxy: isProxy$1 } = is;
const { URL: URL$2 } = url$1;

const clientRequestListeners = new Set();
const fileReadListeners = new Set();
const processOnListeners = new Set();
const originalHttpRequest = http.request;
const originalHttpGet = http.get;
const originalHttpsRequest = https.request;
const originalHttpsGet = https.get;
const originalProcessOn = process.on;
const originalReadStreamRead = fs.ReadStream.prototype._read;
const originalReadFile = fs.readFile;
const originalReadFileSync = fs.readFileSync;

var intercept = {
  interceptClientRequest,
  interceptFileRead,
  interceptProcessOn
};

/**
 * Listen for file system reads and report
 *
 * @param { (filePath: string) => void } fn
 * @returns { () => void }
 */
function interceptFileRead(fn) {
  if (!isProxy$1(fs.readFile)) {
    // Proxy ReadStream private method to work around patching by graceful-fs
    const ReadStream = fs.ReadStream.prototype;

    ReadStream._read = new Proxy(ReadStream._read, {
      apply(target, ctx, args) {
        if (notifyListeners(fileReadListeners, ctx.path) === false) {
          return;
        }
        return Reflect.apply(target, ctx, args);
      }
    });

    for (const method of ['readFile', 'readFileSync']) {
      fs[method] = new Proxy(fs[method], {
        apply(target, ctx, args) {
          if (notifyListeners(fileReadListeners, args[0]) === false) {
            return;
          }
          return Reflect.apply(target, ctx, args);
        }
      });
    }
  }

  fileReadListeners.add(fn);
  return restoreFileRead.bind(null, fn);
}

/**
 * Restore unproxied file reading behaviour
 *
 * @param { (filePath: string) => void } fn
 */
function restoreFileRead(fn) {
  fileReadListeners.delete(fn);
  if (!fileReadListeners.size) {
    fs.ReadStream.prototype._read = originalReadStreamRead;
    fs.readFile = originalReadFile;
    fs.readFileSync = originalReadFileSync;
  }
}

/**
 * Listen for client requests
 *
 * @param { (url: URL) => boolean } fn
 * @returns { () => void }
 */
function interceptClientRequest(fn) {
  if (!isProxy$1(http.request)) {
    http.request = new Proxy(http.request, {
      apply: clientRequestApplyTrap('http')
    });
    http.get = new Proxy(http.get, { apply: clientRequestApplyTrap('http') });
    https.request = new Proxy(https.request, {
      apply: clientRequestApplyTrap('https')
    });
    https.get = new Proxy(https.get, {
      apply: clientRequestApplyTrap('https')
    });
  }

  clientRequestListeners.add(fn);
  return restoreClientRequest.bind(null, fn);
}

/**
 * Restore unproxied client request behaviour
 *
 * @param { (url: URL) => boolean } fn
 */
function restoreClientRequest(fn) {
  clientRequestListeners.delete(fn);
  if (!clientRequestListeners.size) {
    http.request = originalHttpRequest;
    http.get = originalHttpGet;
    https.request = originalHttpsRequest;
    https.get = originalHttpsGet;
  }
}

/**
 * Create client request Proxy apply trap for 'protocol'
 *
 * @param { string } protocol
 * @returns { ClientRequest }
 */
function clientRequestApplyTrap(protocol) {
  return function apply(target, ctx, args) {
    const isString = typeof args[0] === 'string';
    const url = new URL$2(
      isString ? args[0] : getHrefFromRequestOptions(args[0], protocol)
    );

    // TODO: pass method/headers
    if (notifyListeners(clientRequestListeners, url) === false) {
      return;
    }

    if (isLocalhost$1(url.hostname)) {
      // Force to http
      url.protocol = 'http:';
      target =
        target === originalHttpsGet || target === originalHttpGet
          ? originalHttpGet
          : originalHttpRequest;
    }
    if (isString) {
      args[0] = url.href;
    } else {
      args[0].protocol = url.protocol;
      args[0].host = url.host;
      args[0].hostname = url.hostname;
      args[0].port = url.port;
      args[0].path = `${url.href.replace(url.origin, '')}`;
      args[0].href = url.href;
      // Force http agent
      if (args[0].agent && args[0].agent.protocol === 'https:') {
        args[0].agent = new http.Agent(args[0].agent.options || {});
      }
    }

    return Reflect.apply(target, ctx, args);
  };
}

/**
 * Retrieve href from 'options'
 *
 * @param { object } options
 * @param { string } options.href
 * @param { string } options.host
 * @param { string } options.path
 * @param { string } options.port
 * @param { string } protocol
 * @returns { string }
 */
function getHrefFromRequestOptions(options, protocol) {
  if (options.href) {
    return options.href;
  }

  let { host = 'localhost', path, port } = options;

  if (!host.includes(':') && port != null) {
    host += `:${port}`;
  }

  return `${protocol}://${host}${path}`;
}

/**
 * Listen for process event registration
 *
 * @param { (event: string, () => void) => void } fn
 * @returns { () => void }
 */
function interceptProcessOn(fn) {
  if (!isProxy$1(process.on)) {
    process.on = new Proxy(process.on, {
      apply: (target, ctx, args) => {
        if (notifyListeners(processOnListeners, ...args) === false) {
          return;
        }
        return Reflect.apply(target, ctx, args);
      }
    });
  }

  processOnListeners.add(fn);
  return restoreProcessOn.bind(null, fn);
}

/**
 * Restore unproxied process event registration
 *
 * @param { (event: string, () => void) => void } fn
 */
function restoreProcessOn(fn) {
  processOnListeners.delete(fn);
  if (!processOnListeners.size) {
    process.on = originalProcessOn;
  }
}

/**
 * Notify 'listeners' with 'args'
 *
 * @param { Set<string> } listeners
 * @param { ...any } args
 */
function notifyListeners(listeners, ...args) {
  for (const listener of listeners) {
    if (listener(...args) === false) {
      return false;
    }
  }
}

var scripts = {
  concatScripts,
  getProcessEnvString,
  hashScript
};

/**
 * Retrieve process.env polyfill
 *
 * @returns { string }
 */
function getProcessEnvString() {
  return `window.process=window.process||{env:{}};window.process.env.NODE_ENV="${process
    .env.NODE_ENV || 'development'}";`;
}

/**
 * Concatenate multiple "scripts" into a single string
 *
 * @param { Array<string> } scripts
 * @return { string }
 */
function concatScripts(scripts) {
  return scripts.filter((script) => !!script).join('\n');
}

/**
 * Retrieve sha256 hash of "script"
 *
 * @param { string } script
 * @returns { string }
 */
function hashScript(script) {
  return crypto
    .createHash('sha256')
    .update(script)
    .digest('base64');
}

/**
 * @typedef { import("http").ClientRequest } ClientRequest
 * @typedef { import("http").ServerResponse } ServerResponse
 * @typedef { import("../push-events/index.js").PushEvent } PushEvent
 * @typedef { import("../push-events/index.js").pushEvent } pushEvent
 * @typedef { import("../push-events/index.js").PushStream } PushStream
 */
/**
 * @typedef { object } MockRequest
 * @property { string } url
 * @property { string } filePath
 * @property { boolean } [ignoreSearch]
 */
/**
 * @typedef { object } MockResponse
 * @property { object } headers
 * @property { string | object } body
 */
/**
 * @typedef { object } MockPushStream
 * @property { string } url
 * @property { string } filePath
 * @property { string } type
 * @property { boolean } [ignoreSearch]
 * @property { string } [protocol]
 */
/**
 * @typedef { object } MockPushEvent
 * @property { string } name
 * @property { string | object } [message]
 * @property { Array<MockPushEvent> } [sequence]
 * @property { object } [options]
 * @property { number } [options.delay]
 * @property { string } [options.event]
 * @property { string } [options.id]
 */

const { error: error$3, info: info$2, noisyInfo: noisyInfo$2 } = log;
const { getUrl: getUrl$2, getUrlCacheKey: getUrlCacheKey$2, isWebSocketUrl: isWebSocketUrl$2 } = url;
const { isInvalidFilePath: isInvalidFilePath$1, isJsonFilePath: isJsonFilePath$1 } = is;


const debug$2 = debug$a('dvlp:mock');
const { getProjectPath: getProjectPath$1 } = file;

const { interceptClientRequest: interceptClientRequest$1 } = intercept;





const mockClient =
  "!function(){const e=/wss?:/,n=new Map($MOCKS),t={};for(const e of n.values())for(const n in e)e[n].events&&(t[e[n].url]=e[n].events);window.XMLHttpRequest.prototype.open=new Proxy(window.XMLHttpRequest.prototype.open,{apply:(e,n,t)=>Reflect.apply(e,n,l(t,1))}),window.fetch=new Proxy(window.fetch,{apply:(e,n,t)=>Reflect.apply(e,n,l(t,0))}),window.EventSource=new Proxy(window.EventSource,{construct:(e,n)=>Reflect.construct(e,l(n,0))}),window.WebSocket=new Proxy(window.WebSocket,{construct:(e,n)=>Reflect.construct(e,l(n,0))});const o=\"color: hotpink; font-size: 14px; font-weight: bold; text-decoration: underline dotted\",c=\"color: hotpink; font-weight: bold; text-decoration: underline dotted\",s=\"color: grey; font-weight: bold\",i=\"color: grey\",r=\"color: grey; font-style: italic\",p=\"color: grey; font-style: italic; font-weight: bold\";function l(t,o){const c=new URL(t[o],location.origin);if(function(e){const t=function(e){let n=`${e.host}${e.pathname}`;n.endsWith(\"/\")&&(n=n.slice(0,-1));return n}(e),o=n.get(t);if(!o)return;if(!e.search)return o.default;if(e.search in o)return o[e.search];if(o.default&&o.default.ignoreSearch)return o.default}(c)){const n=new URL(`${e.test(c.protocol)?\"ws:\":location.protocol}${location.host}${location.pathname}`);n.searchParams.append(\"dvlpmock\",encodeURIComponent(c.href)),t[o]=n.href}return t}window.dvlp={events:t,pushEvent(e,n){fetch(\"/dvlp/push-event\",{method:\"POST\",headers:{Accept:\"application/json\",\"Content-Type\":\"application/json\"},body:JSON.stringify({stream:e,event:n})})}},Object.defineProperty(window.dvlp,\"help\",{get(){console.log(\"\\n%cWelcome to the dvlp client command-line interface!\\n\\n%cUsage:%c dvlp.<command>\\n\\n%cCommands:\\n%c  pushEvent%c - trigger EventSource/WebSocket event when mocking (%cdvlp.pushEvent.help%c for more)\\n\",o,c,p,c,s,i,r,i)}}),Object.defineProperty(window.dvlp.pushEvent,\"help\",{get(){console.log('\\n%cPushEvent: trigger EventSource/WebSocket event when mocking\\n\\n%cUsage:%c dvlp.pushEvent(stream:%c string%c, event:%c string|object%c)\\n\\n%cArguments:\\n%c  stream:%c string%c - the mocked EventSource/WebSocket connection endpoint\\n%c  event:%c string%c - the \"name\" of a mocked event\\n%c  event:%c {message, options}%c - event object\\n%c    message:%c string|object%c - the event payload as string or JSON object\\n%c    [options]:%c {event: string, id: string}%c - optional EventSource \"event\" name and \"id\"\\n\\n%cExamples:\\n%c  // Push mock event named \"newsfeed with 9 items\"\\n  dvlp.pushEvent(\"http://localhost:8000/feed\", \"newsfeed with 9 items\");\\n\\n  // Push event with string message payload\\n  dvlp.pushEvent(\"http://localhost:8000/feed\", {message: \"some news\"});\\n\\n  // Push event with object message payload and EventSource event name\\n  dvlp.pushEvent(\"http://localhost:8000/feed\", {\\n    message: {item: \"some news\"},\\n    options: {event: \"news\"}\\n  });\\n\\n%cSee:%c https://github.com/popeindustries/dvlp#mocking\\n\\n%cCurrently mocked events:\\n%c'+Object.keys(t).map(e=>`  ${e}\\n    ${t[e].join(\"\\n    \")}`).join(\"\\n\")+\"\\n\",o,c,p,r,p,r,p,c,p,r,i,p,r,i,p,r,i,p,r,i,p,r,i,c,r,c,s,c,p)}})}();" ||
  fs.readFileSync(path.resolve(__dirname, 'mock-client.js'), 'utf8');

var mock = class Mock {
  /**
   * Constructor
   *
   * @param { string | Array<string> } [filePaths]
   */
  constructor(filePaths) {
    this.cache = new Map();
    this.clean = this.clean.bind(this);
    this._uninterceptClientRequest;

    if (filePaths) {
      this.load(filePaths);
    }

    // Client mocking only relevant for loaded mocks,
    // so safe to serialize after call to this.load()
    this.client = mockClient.replace(
      /\$MOCKS/g,
      JSON.stringify(
        [...this.cache].map(([key, value]) => {
          const clientValue = {};

          for (const entry in value) {
            clientValue[entry] = {
              ignoreSearch: value[entry].ignoreSearch,
              url: value[entry].url.href
            };
            if (value[entry].events) {
              clientValue[entry].events = Object.keys(value[entry].events);
            }
          }

          return [key, clientValue];
        })
      )
    );
  }

  /**
   * Add new mock for 'res'
   *
   * @param { string | MockRequest } req
   * @param { MockResponse } res
   * @param { boolean } once
   */
  addResponse(req, res, once) {
    if (!res.body) {
      res = { body: res, headers: {} };
    }

    const url = getUrl$2(req);
    const key = getUrlCacheKey$2(url);
    const ignoreSearch = req.ignoreSearch || false;
    const search = ignoreSearch || !url.search ? 'default' : url.search;
    const type =
      typeof res.body === 'string'
        ? isInvalidFilePath$1(res.body)
          ? 'html'
          : 'file'
        : 'json';
    const filePath = req.filePath || path.join(process.cwd(), 'mock');
    // Allow multiple (subkeyed on search)
    const mock = this.cache.get(key) || {};

    mock[search] = {
      key: search,
      filePath,
      url,
      ignoreSearch,
      once,
      type,
      response: res
    };

    if (!this._uninterceptClientRequest) {
      this._initRequestInterception();
    }
    this.cache.set(key, mock);
    debug$2(`adding mocked "${url.href}"`);
  }

  /**
   * Add new push mock for 'events'
   *
   * @param { string | MockPushStream } stream
   * @param { MockPushEvent | Array<MockPushEvent> } events
   */
  addPushEvents(stream, events) {
    if (!Array.isArray(events)) {
      events = [events];
    }
    const url = getUrl$2(stream);
    const key = getUrlCacheKey$2(url);
    const ignoreSearch = stream.ignoreSearch || false;
    const search = ignoreSearch || !url.search ? 'default' : url.search;
    const type = isWebSocketUrl$2(url) ? 'ws' : 'es';
    // Default to socket.io protocol for ws
    const protocol = stream.protocol || type;
    const filePath = stream.filePath || path.join(process.cwd(), 'mock');
    // Allow multiple (subkeyed on search)
    const mock = this.cache.get(key) || {};

    mock[search] = mock[search] || {
      key: search,
      filePath,
      url,
      ignoreSearch,
      type,
      protocol,
      events: {}
    };

    for (const event of events) {
      // Ignore events without a name
      if (event.name) {
        // Set options for ws
        if (type === 'ws') {
          const options = event.options || {};

          options.protocol = protocol;
          event.options = options;
        }
        // Overwrite existing
        mock[search].events[event.name] = event;
      }
    }

    this.cache.set(key, mock);
    debug$2(`adding mocked stream "${url.href}"`);
  }

  /**
   * Load mock files from disk
   *
   * @param { string | Array<string> } filePaths
   */
  load(filePaths) {
    if (!Array.isArray(filePaths)) {
      filePaths = [filePaths];
    }

    for (let filePath of filePaths) {
      filePath = path.resolve(filePath);

      try {
        const stat = fs.statSync(filePath);

        if (stat.isDirectory()) {
          this.loadDirectory(filePath);
        } else {
          this.loadFile(filePath);
        }
      } catch (err) {
        error$3(`unable to find mock file ${filePath}`);
      }
    }
  }

  /**
   * Match and handle mock response for 'req'
   * Will respond if 'res' passed
   *
   * @param { string | ClientRequest } req
   * @param { ServerResponse } [res]
   * @returns { boolean | object | undefined }
   */
  matchResponse(req, res) {
    const mock = this.getMock(req);

    if (!mock) {
      return false;
    }

    if (!res) {
      return mock;
    }

    const {
      once,
      response: { hang, error, missing, offline }
    } = mock;

    if (once) {
      this.remove(req);
    }

    // Handle special status
    if (hang) {
      return;
    } else if (error || missing) {
      const statusCode = error ? 500 : 404;
      const body = error ? 'error' : 'missing';

      res.writeHead(statusCode);
      res.end(body);
      return;
    } else if (offline) {
      req.socket.destroy();
      return;
    }

    debug$2(`sending mocked "${mock.url.href}"`);

    const {
      filePath,
      response: { body, headers = {} },
      type
    } = mock;
    let content = body;

    switch (type) {
      case 'file':
        // Body is path to file (relative to mock file)
        send(
          { url: req, headers: {} },
          path.resolve(path.dirname(filePath), body),
          {
            cacheControl: true,
            dotfiles: 'allow',
            maxAge: config.maxAge
          }
        ).pipe(res);
        return;
      case 'json':
        content = JSON.stringify(body);
        break;
    }

    res.writeHead(200, {
      // Allow type to be overwritten
      'Content-Type': mime.getType(type),
      ...headers,
      'Content-Length': Buffer.byteLength(content),
      // Overwrite Date
      Date: new Date().toUTCString()
    });
    res.end(content);

    info$2(
      `${stopwatch.stop(
        res.url,
        true,
        true
      )} handled mocked request for ${chalk.green(
        decodeURIComponent(req.url ? req.url : req)
      )}`
    );

    return true;
  }

  /**
   * Match and handle mock push event for 'stream'
   *
   * @param { string | MockPushStream } stream
   * @param { string } name
   * @param { (stream: string | PushStream, event: PushEvent) => void } push
   * @returns { boolean }
   */
  matchPushEvent(stream, name, push) {
    const mock = this.getMock(stream);
    let sequence;

    if (!mock) {
      return false;
    }

    const event = mock.events[name];

    if (name !== 'connect' || (name === 'connect' && event !== undefined)) {
      if (event === undefined) {
        return false;
      }

      sequence =
        event.sequence === undefined
          ? [{ message: event.message, options: event.options }]
          : event.sequence;
    } else {
      sequence = Object.keys(mock.events).reduce((sequence, name) => {
        const event = mock.events[name];

        if (event.connect) {
          event.sequence === undefined
            ? sequence.push({ message: event.message, options: event.options })
            : sequence.push(...event.sequence);
        }

        return sequence;
      }, []);

      if (sequence.length === 0) {
        return false;
      }
    }

    stopwatch.start(name);
    triggerEventSequence(stream, sequence, push).then(() => {
      noisyInfo$2(
        `${stopwatch.stop(
          name,
          true,
          true
        )} triggered mocked push event ${chalk.green(name)}`
      );
    });

    return true;
  }

  /**
   * Determine if 'url' matches cached mock
   *
   * @param { URL } url
   * @returns { boolean }
   */
  hasMatch(url) {
    return this.getMock(url) !== undefined;
  }

  /**
   * Remove existing mock
   *
   * @param { string | ClientRequest | MockPushStream } reqOrStream
   */
  remove(reqOrStream) {
    const { key, url } = this.getMock(reqOrStream);
    const cacheKey = getUrlCacheKey$2(url);
    const mock = this.cache.get(cacheKey);

    delete mock[key];

    if (!Object.keys(mock).length) {
      this.cache.delete(cacheKey);
      if (!this.cache.size && this._uninterceptClientRequest) {
        this._uninterceptClientRequest();
      }
    }
  }

  /**
   * Clear all mocks
   */
  clean() {
    this.cache.clear();
    this._uninterceptClientRequest && this._uninterceptClientRequest();
  }

  /**
   * Initialize request interception
   *
   * @private
   */
  _initRequestInterception() {
    this._uninterceptClientRequest = interceptClientRequest$1((url) => {
      if (this.hasMatch(url.href)) {
        url.searchParams.append('dvlpmock', encodeURIComponent(url.href));
        // Reroute to active server
        url.host = `localhost:${config.activePort}`;
      }
    });
  }

  /**
   * Retrieve mock
   *
   * @param { string | ClientRequest | MockPushStream } reqOrStream
   * @returns { object }
   * @private
   */
  getMock(reqOrStream) {
    const url = getUrl$2(reqOrStream);
    const key = getUrlCacheKey$2(url);
    const mock = this.cache.get(key);

    if (!mock) {
      return;
    } else if (!url.search) {
      return mock.default;
    } else if (url.search in mock) {
      return mock[url.search];
    } else {
      if (mock.default && mock.default.ignoreSearch) {
        return mock.default;
      }
    }
  }

  /**
   * Load directory at 'dirpath'
   *
   * @param { string } dirpath
   * @private
   */
  loadDirectory(dirpath) {
    fs.readdirSync(dirpath).forEach((filePath) => {
      this.load(path.join(dirpath, filePath));
    });
  }

  /**
   * Load file at 'filePath'
   *
   * @param { string } filePath
   * @private
   */
  loadFile(filePath) {
    if (!isJsonFilePath$1(filePath)) {
      return;
    }

    try {
      let mocks = JSON.parse(fs.readFileSync(filePath, 'utf8'));
      let count = 0;
      let type = '';

      if (!Array.isArray(mocks)) {
        mocks = [mocks];
      }

      for (const mock of mocks) {
        if (!isMockFile(mock)) {
          return;
        } else if (!isValidSchema(mock)) {
          return error$3(`invalid mock format for ${filePath}`);
        }

        const { request, response, stream, events } = mock;
        if (request && response) {
          request.filePath = filePath;
          this.addResponse(request, response, false);
          type = type === 'event' ? 'event/response' : 'response';
          count++;
        } else if (stream && events) {
          stream.filePath = filePath;
          this.addPushEvents(stream, events);
          type = type === 'response' ? 'event/response' : 'event';
          count += events.length;
        }
      }

      info$2(
        `${chalk.green('✔')} loaded ${count} mock ${type}${
          count > 1 ? 's' : ''
        } from ${chalk.green(getProjectPath$1(filePath))}`
      );
    } catch (err) {
      error$3(err);
    }
  }
};

function isMockFile(json) {
  return 'request' in json || 'response' in json || 'stream' in json;
}

/**
 * Validate that 'json' is correct format
 *
 * @param { object } json
 * @returns { boolean }
 * @private
 */
function isValidSchema(json) {
  return (
    ('request' in json &&
      'url' in json.request &&
      'response' in json &&
      'body' in json.response) ||
    ('stream' in json && 'url' in json.stream && 'events' in json)
  );
}

/**
 * Match and handle mock push event for 'stream'
 *
 * @param { string | MockPushStream } stream
 * @param { Array<MockPushEvent> } sequence
 * @param { (stream: string | PushStream, event: PushEvent) => void } push
 * @returns { void }
 * @private
 */
async function triggerEventSequence(stream, sequence, push) {
  for (const event of sequence) {
    const { message, options: { delay = 0, ...options } = {} } = event;

    await sleep(delay);
    push(stream, { message, options });
  }
}

/**
 * Sleep for 'duration' milliseconds
 *
 * @param { number } duration
 * @returns { Promise<void> }
 * @private
 */
function sleep(duration) {
  return new Promise((resolve) => setTimeout(resolve, duration));
}

/**
 * @typedef { import("http").ClientRequest } ClientRequest
 * @typedef { import("http").ServerResponse } ServerResponse
 */

const { bundle: bundle$1, resolveModuleId: resolveModuleId$1 } = bundler_1;
const { getAbsoluteProjectPath, getProjectPath: getProjectPath$2 } = file;
const {
  isCssRequest: isCssRequest$1,
  isHtmlRequest: isHtmlRequest$1,
  isJsRequest: isJsRequest$1,
  isNodeModuleFilePath: isNodeModuleFilePath$2,
  isModule: isModule$1,
  isModuleBundlerFilePath: isModuleBundlerFilePath$1,
  isRelativeFilePath: isRelativeFilePath$3
} = is;
const { warn: warn$1, WARN_BARE_IMPORT: WARN_BARE_IMPORT$1 } = log;

const debug$3 = debug$a('dvlp:patch');
const { filePathToUrl: filePathToUrl$1 } = url;

const { resolve: resolve$2 } = resolver;
const { unzipSync } = zlib;

const RE_CLOSE_BODY_TAG = /<\/body>/i;
const RE_IMPORT = /(import\b[^'"]+['"]|export\b.+from\s+['"])([^'"\n]+)(['"])/gm;
const RE_NONCE_SHA = /nonce-|sha\d{3}-/;
const RE_OPEN_HEAD_TAG = /<head>/i;

var patch = {
  patchResponse
};

/**
 * Patch response body.
 * Injects reload script into html response,
 * and resolves import ids for js/css response.
 *
 * @param { string } filePath
 * @param { ClientRequest } req
 * @param { ServerResponse } res
 * @param { object } [options]
 * @param { Array<string> } [options.directories]
 * @param { object } [options.rollupConfig]
 * @param { object } [options.footerScript]
 * @param { string } options.footerScript.hash
 * @param { string } options.footerScript.string
 * @param { string } options.footerScript.url
 * @param { object } [options.headerScript]
 * @param { string } options.headerScript.hash
 * @param { string } options.headerScript.string
 * @param { string } options.headerScript.url
 */
function patchResponse(
  filePath,
  req,
  res,
  { rollupConfig, footerScript, headerScript } = {}
) {
  // req.filepath set after file.find(), filepath passed if cached
  filePath = req.filePath || filePath || req.url;
  debug$3(`patching response for "${getProjectPath$2(filePath)}"`);
  proxySetHeader(res, disableHeaders.bind(disableHeaders, res, req.url));
  if (isHtmlRequest$1(req)) {
    if (footerScript || headerScript) {
      const urls = [];
      const hashes = [];
      const scripts = {};

      if (footerScript) {
        if (footerScript.url) {
          urls.push(footerScript.url);
        }
        if (footerScript.hash) {
          hashes.push(footerScript.hash);
        }
        scripts.footer = footerScript.string;
      }
      if (headerScript) {
        if (headerScript.url) {
          urls.push(headerScript.url);
        }
        if (headerScript.hash) {
          hashes.push(headerScript.hash);
        }
        scripts.header = headerScript.string;
      }
      proxySetHeader(res, injectCSPHeader.bind(injectCSPHeader, urls, hashes));
      proxyBodyWrite(res, injectScripts.bind(injectScripts, scripts));
    }
  } else if (isCssRequest$1(req)) {
    // Disable gzip
    proxyBodyWrite(res, (data) => data);
  } else if (isJsRequest$1(req)) {
    proxyBodyWrite(
      res,
      rewriteImports.bind(rewriteImports, filePath, rollupConfig)
    );
  }
}

/**
 * Disable Cache-Control, Content-Encoding headers
 *
 * @param { ServerResponse } res
 * @param { string } url
 * @param { string } key
 * @param { string } value
 * @returns { string }
 */
function disableHeaders(res, url, key, value) {
  const lcKey = key.toLowerCase();

  if (
    lcKey === 'cache-control' &&
    !isNodeModuleFilePath$2(url) &&
    !isModuleBundlerFilePath$1(url)
  ) {
    value = 'no-cache, dvlp-disabled';
  }
  if (lcKey === 'content-encoding') {
    res.encoding = value;
    value = undefined;
  }

  return value;
}

/**
 * Inject header/footer script tags into 'data'
 *
 * @param { { footer: string, header: string } } scripts
 * @param { string } data
 * @returns { string }
 */
function injectScripts(scripts, data) {
  const { footer, header } = scripts;

  if (header && RE_OPEN_HEAD_TAG.test(data)) {
    debug$3('injecting header script');
    data = data.replace(RE_OPEN_HEAD_TAG, `<head>\n<script>${header}</script>`);
  }
  if (footer && RE_CLOSE_BODY_TAG.test(data)) {
    debug$3('injecting footer script');
    data = data.replace(
      RE_CLOSE_BODY_TAG,
      `<script>${footer}</script>\n</body>`
    );
  }

  return data;
}

/**
 * Inject CSP headers allowing inline script tag
 *
 * @param { Array<string> } urls
 * @param { Array<string> } hashes
 * @param { string } key
 * @param { string } value
 * @returns { string }
 */
function injectCSPHeader(urls, hashes, key, value) {
  const lcKey = key.toLowerCase();

  if (
    lcKey === 'content-security-policy-report-only' ||
    lcKey === 'content-security-policy'
  ) {
    urls = urls.join(' ');
    hashes = hashes.map((hash) => `'sha256-${hash}'`).join(' ');
    const rules = value
      .split(';')
      .map((ruleString) => ruleString.trim())
      .reduce((rules, ruleString) => {
        const firstSpace = ruleString.indexOf(' ');

        rules[ruleString.slice(0, firstSpace)] = ruleString.slice(
          firstSpace + 1
        );
        return rules;
      }, {});

    if (rules['connect-src']) {
      rules['connect-src'] = `${rules['connect-src']} ${urls}`;
    } else {
      rules['connect-src'] = urls;
    }
    if (
      rules['script-src'] &&
      (RE_NONCE_SHA.test(value) || !value.includes('unsafe-inline'))
    ) {
      rules['script-src'] = `${rules['script-src']} ${hashes}`;
    }

    value = Object.keys(rules).reduce((value, name) => {
      value += `${name} ${rules[name]}; `;
      return value;
    }, '');
  }

  return value;
}

/**
 * Rewrite bare import references in 'data'
 *
 * @param { string } filePath
 * @param { object } rollupConfig
 * @param { Buffer | string } data
 * @returns { Buffer | string }
 */
function rewriteImports(filePath, rollupConfig, data) {
  const asBuffer = data instanceof Buffer;
  const projectFilePath = getProjectPath$2(filePath);
  let code = asBuffer ? data.toString() : data;
  let match;

  if (!RE_IMPORT.test(code)) {
    debug$3(`no imports to rewrite in "${projectFilePath}"`);
    return data;
  }

  RE_IMPORT.lastIndex = 0;
  while ((match = RE_IMPORT.exec(code))) {
    const [context, pre, id, post] = match;
    const importPath = resolve$2(id, getAbsoluteProjectPath(filePath));

    if (importPath) {
      let newId = '';

      // Bundle if in node_modules and not an es module
      if (isNodeModuleFilePath$2(importPath) && !isModule$1(importPath)) {
        const resolvedId = resolveModuleId$1(id, importPath);

        // Trigger bundling in background while waiting for eventual request
        bundle$1(resolvedId, id, rollupConfig);
        newId = `/${path.join(config.bundleDirName, resolvedId)}`;
        warn$1(WARN_BARE_IMPORT$1, id);
      } else {
        // Don't rewrite if no change after resolving
        newId =
          isRelativeFilePath$3(id) &&
          path.join(path.dirname(filePath), id) === importPath
            ? id
            : `/${getProjectPath$2(importPath)}`;
      }

      newId = filePathToUrl$1(newId);
      debug$3(`rewrote import id from "${id}" to "${newId}"`);
      code = code.replace(context, `${pre}${newId}${post}`);
    } else {
      warn$1(`⚠️  unable to resolve path for "${id}" from "${projectFilePath}"`);
    }
  }

  if (asBuffer) {
    code = Buffer.from(code);
  }

  return code;
}

/**
 * Proxy set header for 'res', performing 'action' on write()/end()
 *
 * @param { ServerResponse } res
 * @param { function } action
 */
function proxySetHeader(res, action) {
  res.setHeader = new Proxy(res.setHeader, {
    apply(target, ctx, args) {
      let [key, value] = args;

      value = action(key, value);

      if (value) {
        return Reflect.apply(target, ctx, [key, value]);
      }
    }
  });

  res.writeHead = new Proxy(res.writeHead, {
    apply(target, ctx, args) {
      // First argument is always statusCode
      if (args.length > 1) {
        const headers = args[args.length - 1];

        for (const key in headers) {
          const value = action(key, headers[key]);

          if (value) {
            headers[key] = value;
          }
        }
      }

      return Reflect.apply(target, ctx, args);
    }
  });
}

/**
 * Proxy body write for 'res', performing 'action' on write()/end()
 *
 * @param { ServerResponse } res
 * @param { function } action
 */
function proxyBodyWrite(res, action) {
  const originalSetHeader = res.setHeader;
  let buffer;

  // Proxy write() to buffer streaming response
  res.write = new Proxy(res.write, {
    apply(target, ctx, args) {
      let [chunk] = args;

      if (typeof chunk === 'string') {
        chunk = Buffer.from(chunk);
      }
      buffer = Buffer.concat([buffer || Buffer.from(''), chunk]);
      return;
    }
  });

  // Proxy end() to intercept response body
  res.end = new Proxy(res.end, {
    apply(target, ctx, args) {
      let [data = buffer, ...extraArgs] = args;
      let size = 0;

      if (data) {
        if (Buffer.isBuffer(data)) {
          if (res.encoding === 'gzip') {
            data = unzipSync(data);
          }
          data = data.toString();
        }
        data = action(data);
        size = Buffer.byteLength(data);
      }

      if (!res.headersSent) {
        if (size) {
          debug$3(`setting Content-Length to ${size}`);
          originalSetHeader.call(res, 'Content-Length', size);
        }
      }

      return Reflect.apply(target, ctx, [data, ...extraArgs]);
    }
  });

  // Prevent setting of Content-Length
  res.setHeader = new Proxy(res.setHeader, {
    apply(target, ctx, args) {
      let [key, value] = args;

      if (key.toLowerCase() === 'content-length') {
        debug$3(`prevented setting Content-Length to ${value}`);
        return;
      }

      return Reflect.apply(target, ctx, [key, value]);
    }
  });

  // Prevent setting of Content-Length
  res.writeHead = new Proxy(res.writeHead, {
    apply(target, ctx, args) {
      // First argument is always statusCode
      if (args.length > 1) {
        for (const key in args[args.length - 1]) {
          if (key.toLowerCase() === 'content-length') {
            debug$3(
              `prevented setting Content-Length to ${
                args[args.length - 1][key]
              }`
            );
            delete args[args.length - 1][key];
          }
        }
      }

      return Reflect.apply(target, ctx, args);
    }
  });
}

/**
 * @typedef { import("http").ServerResponse } ServerResponse
 */

const { getProjectPath: getProjectPath$3, getTypeFromPath } = file;

const debug$4 = debug$a('dvlp:transpile');
const { error: error$4, info: info$3 } = log;



/**
 * Transpile content for filePath
 *
 * @param { string } filePath
 * @param { ServerResponse } res
 * @param { object } state
 * @param { Map<string, string> } state.transpilerCache
 * @param { string } state.lastChanged
 * @param { (filePath: string, isServer: boolean) => Promise<string> | string | undefined } state.transpiler
 * @returns { Promise<void> }
 */
var transpile = async function transpile(filePath, res, state) {
  const { transpilerCache, lastChanged, transpiler } = state;
  const relativeFilePath = getProjectPath$3(filePath);
  // Dependencies that are concatenated during transpile aren't cached,
  // but they are watched when read from file system during transpilation,
  // so transpile again if same type
  const lastChangedIsDependency =
    !!lastChanged &&
    !transpilerCache.has(lastChanged) &&
    getTypeFromPath(lastChanged) === getTypeFromPath(filePath);
  let content = transpilerCache.get(filePath);
  let transpiled = false;

  if (lastChangedIsDependency || lastChanged === filePath || !content) {
    try {
      content = await transpiler(filePath, false);
      if (content) {
        transpiled = true;
        transpilerCache.set(filePath, content);
      }
    } catch (err) {
      debug$4(`error transpiling "${relativeFilePath}"`);
      res.writeHead(500);
      res.end(err.message);
      error$4(err);
      return;
    }
  }

  if (content) {
    debug$4(
      `${
        transpiled ? 'transpiled content for' : 'skipping transpile for'
      } "${relativeFilePath}"`
    );
    res.transpiled = true;
    res.writeHead(200, {
      'Content-Length': Buffer.byteLength(content),
      'Cache-Control': 'max-age=0',
      'Content-Type': mime.getType(getTypeFromPath(filePath) || filePath)
    });
    res.end(content);

    info$3(
      `${stopwatch.stop(
        res.url,
        true,
        true
      )} handled transpiled request for ${chalk.green(res.url)}`
    );
  }
};

/**
 * @typedef { object } watcher
 * @property { (filePath: string) => void } add
 * @property { () => void } close
 */

const debug$5 = debug$a('dvlp:watch');
const { FSWatcher } = chokidar;
const { getProjectPath: getProjectPath$4 } = file;


// Hidden, node_modules/*, .dvlp/*
const RE_IGNORED = /(^|[/\\])\..|node_modules|\.dvlp/i;
const TIMEOUT = 1000;

/**
 * Instantiate a file watcher and begin watching for changes
 *
 * @param { (string) => void } fn
 * @returns { watcher }
 */
var watch = function watch(fn) {
  const watcher = new FSWatcher({
    ignored: RE_IGNORED,
    ignoreInitial: true,
    persistent: true
  });
  const files = new Set();
  let changing = false;

  watcher.on('unlink', (filePath) => {
    debug$5(`unwatching file "${getProjectPath$4(filePath)}"`);
    watcher.unwatch(filePath);
    files.delete(path.resolve(filePath));
  });
  watcher.on('change', (filePath) => {
    if (!changing) {
      // Prevent double change
      setTimeout(() => {
        changing = false;
      }, TIMEOUT);
      changing = true;
      debug$5(`change detected "${getProjectPath$4(filePath)}"`);
      fn(path.resolve(filePath));
    }
  });

  return {
    add(filePath) {
      if (Array.isArray(filePath)) {
        return filePath.forEach(this.add);
      }

      filePath = path.resolve(filePath);
      if (!files.has(filePath) && !RE_IGNORED.test(filePath)) {
        debug$5(`watching file "${getProjectPath$4(filePath)}"`);
        files.add(filePath);
        watcher.add(filePath);
      }
    },
    close() {
      watcher.close();
      files.clear();
    }
  };
};

/**
 * @typedef { import("http").ClientRequest } ClientRequest
 * @typedef { import("../reloader/index.js").reloadServer } reloadServer
 * @typedef { import("http").ServerResponse } ServerResponse
 * @typedef { import("../utils/watch.js").watcher } watcher
 */
/**
 * @typedef { object } appServer
 * @property { () => Promise<void> } destroy
 * @property { number } port
 * @property { () => Promise<void> } restart
 */

const { error: error$5, fatal: fatal$1, info: info$4, noisyInfo: noisyInfo$3 } = log;
const {
  find: find$2,
  getProjectPath: getProjectPath$5,
  getTypeFromRequest,
  importModule
} = file;
const {
  handleMockResponse: handleMockResponse$1,
  handleMockWebSocket: handleMockWebSocket$1,
  handlePushEvent: handlePushEvent$1
} = common;
const {
  interceptFileRead: interceptFileRead$1,
  interceptProcessOn: interceptProcessOn$1
} = intercept;
const {
  isModuleBundlerFilePath: isModuleBundlerFilePath$2,
  isNodeModuleFilePath: isNodeModuleFilePath$3,
  isProjectFilePath: isProjectFilePath$1
} = is;
const { bundle: bundle$2 } = bundler_1;
const {
  concatScripts: concatScripts$1,
  getProcessEnvString: getProcessEnvString$1,
  hashScript: hashScript$1
} = scripts;


const debug$6 = debug$a('dvlp:app');



// Work around rollup-plugin-commonjs require.cache
const moduleCache = module$1._cache;

const { patchResponse: patchResponse$1 } = patch;





const START_TIMEOUT_DURATION = 2000;

const originalCreateServer = http.createServer;
let dvlpModules;

/**
 * Create app server
 *
 * @param { string } mainPath
 * @param { object } [options]
 * @param { string } [options.mockPath]
 * @param { number } [options.port]
 * @param { reloadServer } [options.reloader]
 * @param { object } [options.rollupConfig]
 * @param { (filePath: string, isServer: boolean) => Promise<string> | string | undefined } [options.transpiler]
 * @returns { appServer }
 */
var appServer = async function appServer(
  mainPath,
  { mockPath, port, reloader, rollupConfig, transpiler } = {}
) {
  if (process.env.PORT === undefined) {
    process.env.PORT = port;
  }
  if (dvlpModules === undefined) {
    dvlpModules = Object.keys(moduleCache);
  }

  const main = path.resolve(mainPath);
  const server = new AppServer(
    main,
    reloader,
    rollupConfig,
    transpiler,
    mockPath
  );

  try {
    await server.start();
  } catch (err) {
    return error$5(err);
  }

  info$4(
    `\n  💥 serving ${chalk.green(
      getProjectPath$5(mainPath)
    )} at ${chalk.green.underline(server.origin)}`
  );

  return {
    destroy: server.destroy.bind(server),
    port: server.port,
    restart: server.restart.bind(server)
  };
};

class AppServer {
  /**
   * Constructor
   *
   * @param { string } main
   * @param { reloader } [reloader]
   * @param { object } [rollupConfig]
   * @param { (filePath: string, isServer: boolean) => Promise<string> | string | undefined } [transpiler]
   * @param { string } [mockPath]
   */
  constructor(main, reloader, rollupConfig, transpiler, mockPath) {
    this.appModules = [];
    this.connections = new Map();
    this.exitHandler = null;
    this.lastChanged = '';
    this.main = main;
    this.mocks = mockPath && new mock(mockPath);

    const headerScript = concatScripts$1([
      getProcessEnvString$1(),
      mockPath && this.mocks.client
    ]);

    this.patchResponseOptions = {
      rollupConfig,
      footerScript: {
        hash: reloader && hashScript$1(reloader.client),
        string: reloader && reloader.client,
        url: reloader && reloader.url
      },
      headerScript: {
        hash: hashScript$1(headerScript),
        string: headerScript
      }
    };
    this.origin = '';
    this.port = process.env.PORT;
    this.reloader = reloader;
    this.rollupConfig = rollupConfig;
    this.server = null;
    this.transpiler = transpiler;
    this.transpilerCache = transpiler && new Map();
    this.urlToFilePath = new Map();
    this.watcher = this.createWatcher();

    // Listen for all upcoming file system reads (including require('*'))
    this.unlistenForFileRead = interceptFileRead$1((filePath) => {
      if (isProjectFilePath$1(filePath)) {
        this.watcher.add(filePath);
      }
    });
  }

  /**
   * Create watcher instance and listen for file changes
   *
   * @returns { watcher }
   */
  createWatcher() {
    return watch(async (filePath) => {
      debug$6('restarting server');

      noisyInfo$3(
        `\n  ⏱  ${new Date().toLocaleTimeString()} ${chalk.yellow(
          getProjectPath$5(filePath)
        )}`
      );

      this.lastChanged = filePath;

      try {
        await this.restart();

        for (const [url, fp] of this.urlToFilePath) {
          if (fp === filePath) {
            filePath = url;
            break;
          }
        }

        // Will trigger one of:
        // 1. single css refresh if css and a link.href matches filePath
        // 2. multiple css refreshes if css and no link.href matches filePath (ie. it's a dependency)
        // 3. full page reload
        this.reloader && this.reloader.send(filePath);
      } catch (err) {
        error$5(err);
      }
    });
  }

  /**
   * Start app server
   * Proxies createServer to grab instance and register for events
   *
   * @returns { Promise<void> }
   */
  start() {
    return new Promise((resolve, reject) => {
      // Force reject if never started
      const timeoutID = setTimeout(() => {
        debug$6('server not started after timeout');
        reject(Error('unable to start server'));
      }, START_TIMEOUT_DURATION);
      const instance = this;

      http.createServer = new Proxy(http.createServer, {
        apply(target, ctx, args) {
          // Handler always last arg ('options' as first arg added in 9.6)
          let handler = args[args.length - 1];

          // Wrap request handler (if passed)
          if (handler && typeof handler === 'function') {
            args[args.length - 1] = instance
              .createRequestHandler(handler)
              .bind(instance);
          } else {
            handler = undefined;
          }

          const server = (instance.server = Reflect.apply(target, ctx, args));

          server.timeout = server.keepAliveTimeout = 0;

          server.on('connection', (connection) => {
            const key = `${connection.remoteAddress}:${connection.remotePort}`;

            instance.connections.set(key, connection);
            connection.on('close', () => {
              instance.connections.delete(key);
            });
          });
          server.on('error', (err) => {
            instance.stop();
            reject(err);
          });
          server.on('listening', () => {
            debug$6('server started');
            clearTimeout(timeoutID);
            instance.appModules = getAppModules();
            instance.watcher.add(instance.appModules);
            instance.port = server.address().port;
            instance.origin = `http://localhost:${instance.port}`;
            resolve();
          });
          server.on('upgrade', (req, socket, body) => {
            handleMockWebSocket$1(req, socket, body, instance.mocks);
          });

          // No handler registered so proxy "request" listener
          if (!handler) {
            server.on = new Proxy(server.on, {
              apply(target, ctx, args) {
                const [eventName, listener] = args;

                // Wrap request handler
                if (eventName === 'request') {
                  args[1] = instance
                    .createRequestHandler(listener)
                    .bind(instance);
                }

                Reflect.apply(target, ctx, args);
              }
            });
          }

          // Un-proxy in case more than one server created
          // (assumes first server is application server)
          http.createServer = originalCreateServer;

          return server;
        }
      });

      try {
        // Trigger application bootstrap
        this.startApplication();
      } catch (err) {
        reject(err);
      }
    });
  }

  /**
   * Create request handler wrapper for 'originalRequestHandler'
   *
   * @param { (req: ClientRequest, res: ServerResponse) => void } originalRequestHandler
   * @returns { (req: ClientRequest, res: ServerResponse) => void }
   */
  createRequestHandler(originalRequestHandler) {
    return async function requestHandler(req, res) {
      stopwatch.start(req.url);

      const type = getTypeFromRequest(req);
      let filePath = this.urlToFilePath.get(req.url);

      res.url = req.url;

      if (
        handleMockResponse$1(req, res, this.mocks) ||
        handlePushEvent$1(req, res, this.mocks)
      ) {
        return;
      }

      // Ignore html or uncached or no longer available at previously known path
      if (type !== 'html' && (!filePath || !fs.existsSync(filePath))) {
        filePath = find$2(req, { type });

        if (filePath) {
          this.urlToFilePath.set(req.url, filePath);

          if (isModuleBundlerFilePath$2(filePath)) {
            await bundle$2(path.basename(filePath), undefined, this.rollupConfig);
          }
        } else {
          // File not found. Clear previously known path
          this.urlToFilePath.delete(req.url);
        }
      }

      patchResponse$1(filePath, req, res, this.patchResponseOptions);

      if (filePath) {
        const bundled = isModuleBundlerFilePath$2(filePath);

        // Bundled or node_module esm file
        if (bundled || isNodeModuleFilePath$3(filePath)) {
          info$4(
            `${stopwatch.stop(res.url, true, true)} handled${
              bundled ? ' bundled' : ''
            } request for ${chalk.green(req.url)}`
          );

          return send(req, filePath, {
            cacheControl: true,
            dotfiles: 'allow',
            maxAge: config.maxAge
          }).pipe(res);
        } else if (this.transpiler) {
          // Will respond if transpiler for this type
          await transpile(filePath, res, {
            transpilerCache: this.transpilerCache,
            lastChanged: this.lastChanged,
            transpiler: this.transpiler
          });
        }
      }

      // Not mocked/transpiled/bundled, so pass through request to app
      if (!res.finished) {
        debug$6(`allowing app to handle "${req.url}"`);
        originalRequestHandler(req, res);
      }
    };
  }

  /**
   * Start application
   *
   * @returns { void }
   */
  startApplication() {
    process.on('uncaughtException', this.onUncaught);
    process.on('unhandledRejection', this.onUncaught);
    // Listen for process event registration to capture any exit handlers
    this.unlistenForProcessOn = interceptProcessOn$1((event, handler) => {
      if (event === 'exit' || event === 'beforeExit') {
        this.exitHandler = handler;
        return false;
      }
    });
    importModule(this.main, this.transpiler);
  }

  /**
   * Stop application
   *
   * @returns { void }
   */
  async stopApplication() {
    this.unlistenForProcessOn();
    if (this.exitHandler) {
      await this.exitHandler();
    }
    process.removeListener('uncaughtException', this.onUncaught);
    process.removeListener('unhandledRejection', this.onUncaught);
    clearAppModules(this.appModules, this.main);
  }

  /**
   * Stop running server
   *
   * @returns { Promise<void> }
   */
  stop() {
    return new Promise(async (resolve) => {
      for (const connection of this.connections.values()) {
        connection.destroy();
      }
      this.connections.clear();

      await this.stopApplication();

      if (!this.server) {
        return resolve();
      }

      debug$6('server stopped');
      this.server.removeAllListeners();
      if (!this.server.listening) {
        resolve();
      } else {
        this.server.close(resolve);
      }
    });
  }

  /**
   * Restart running server
   *
   * @returns { Promise<void> }
   */
  async restart() {
    await this.stop();
    return this.start();
  }

  /**
   * Handle uncaughtException & unhandledRejection
   * Logs errors to console
   *
   * @param { Error } err
   * @returns { void }
   */
  onUncaught(err) {
    fatal$1(err);
  }

  /**
   * Destroy instance
   *
   * @returns { Promise<void> }
   */
  destroy() {
    debug$6('destroying');
    this.mocks && this.mocks.clean();
    this.unlistenForFileRead();
    this.watcher.close();
    return this.stop();
  }
}

/**
 * Clear app modules from module cache
 *
 * @param { Array<string> } appModules
 */
function clearAppModules(appModules, main) {
  const mainModule = moduleCache[main];

  // Remove main from parent
  // (No children when bundled)
  if (
    mainModule !== undefined &&
    mainModule.parent !== undefined &&
    mainModule.parent.children !== undefined
  ) {
    const parent = mainModule.parent;
    let i = parent.children.length;

    while (--i) {
      if (parent.children[i].id === mainModule.id) {
        parent.children.splice(i, 1);
      }
    }
  }

  for (const m of appModules) {
    delete moduleCache[m];
  }

  debug$6(`cleared ${appModules.length} app modules from require.cache`);
}

/**
 * Retrieve app modules (excluding node_modules)
 *
 * @returns { Array<string> }
 */
function getAppModules() {
  const modules = Object.keys(moduleCache).filter(
    (m) => !dvlpModules.includes(m) /* && !isNodeModuleFilePath(m) */
  );

  debug$6(`found ${modules.length} app modules`);

  return modules;
}

/**
 * @typedef { object } reloadServer
 * @property { string } client
 * @property { string } clientHash
 * @property { () => Promise<void> } destroy
 * @property { (string, object) => void } send
 * @property { string } url
 */


const debug$7 = debug$a('dvlp:reload');

const { EventSource: EventSource$2 } = fayeWebsocket;


const { getTypeFromPath: getTypeFromPath$1 } = file;

const { noisyInfo: noisyInfo$4 } = log;


const DEFAULT_CLIENT_CONFIG = {
  headers: { 'Access-Control-Allow-Origin': '*' },
  ping: 15,
  retry: 10
};
const DEFAULT_PORT = 3529;
const ENDPOINT = '/dvlpreload';

const reloadClient =
  '!function(){if("undefined"==typeof URL||"undefined"==typeof EventSource)return;const e=new URL(location.origin);let n,o;e.port=$RELOAD_PORT,e.pathname="/dvlpreload",function t(){n=new EventSource(e.href);n.onopen=()=>{o=5};n.onerror=()=>{n.close(),o--&&t()};n.addEventListener("reload",()=>{location.reload()});n.addEventListener("refresh",e=>{const{filePath:n}=JSON.parse(e.data),o=document.querySelectorAll("link"),t={};let r=!1;for(const e of o){const n=new URL(e.href);n.origin.includes("localhost")&&(void 0===e.rel||e.rel.includes("stylesheet"))&&(r=!0,t[n.pathname.slice(1)]=e)}if(!r)return location.reload();if(n in t)t[n].href=n;else for(const e in t)t[e].href=e})}()}();' ||
  fs.readFileSync(path.resolve(__dirname, 'reload-client.js'), 'utf8');

/**
 * Create reload server
 *
 * @returns { Promise<reloadServer> }
 */
var reloader = async function reloadServer() {
  const port = await getPort({
    port: getPort.makeRange(DEFAULT_PORT, DEFAULT_PORT + 100)
  });
  const server = new ReloadServer(port);
  const client = reloadClient.replace(/\$RELOAD_PORT/g, port);

  await server.start();

  return {
    client,
    destroy: server.destroy.bind(server),
    send: server.send.bind(server),
    url: `http://localhost:${server.port}${ENDPOINT}`
  };
};

class ReloadServer {
  /**
   * Constructor
   */
  constructor(port) {
    this.clients = new Set();
    this.port = port;
    this.server = null;
  }

  /**
   * Start server
   *
   * @returns { Promise<void> }
   */
  start() {
    return new Promise((resolve, reject) => {
      this.server = http.createServer(async (req, res) => {
        if (req.url !== ENDPOINT || !EventSource$2.isEventSource(req)) {
          res.writeHead(404);
          res.end();
          return;
        }

        const client = new EventSource$2(req, res, DEFAULT_CLIENT_CONFIG);

        this.clients.add(client);
        debug$7('added reload connection', this.clients.size);

        client.on('close', () => {
          this.clients.delete(client);
          debug$7('removed reload connection', this.clients.size);
        });
      });

      serverDestroy(this.server);
      this.server.timeout = this.server.keepAliveTimeout = 0;
      this.server.unref();
      this.server.on('error', reject);
      this.server.on('listening', resolve);

      this.server.listen(this.port);
    });
  }

  /**
   * Send reload message to clients for changed 'filePath'
   *
   * @param { string } filePath
   * @returns { void }
   */
  send(filePath) {
    const type = getTypeFromPath$1(filePath);
    const event = type === 'css' ? 'refresh' : 'reload';
    const data = JSON.stringify({ type, filePath });

    if (this.clients.size) {
      noisyInfo$4(
        `${chalk.yellow(`⟲  ${event}ing`)} ${this.clients.size} client${
          this.clients.size > 1 ? 's' : ''
        }`
      );

      for (const client of this.clients) {
        client.send(data, { event });
      }
    }
  }

  /**
   * Stop running server
   *
   * @returns { Promise<void> }
   */
  stop() {
    return new Promise((resolve) => {
      if (!this.server) {
        return resolve();
      }

      this.server.removeAllListeners();
      this.server.destroy(() => {
        debug$7('server stopped');
        resolve();
      });
    });
  }

  /**
   * Destroy instance
   *
   * @returns { Promise<void> }
   */
  destroy() {
    debug$7('destroying');
    for (const client of this.clients) {
      client.close();
    }
    this.clients.clear();
    return this.stop();
  }
}

/**
 * @typedef { import("../reloader/index.js").reloadServer } reloadServer
 * @typedef { import("../utils/watch.js").watcher } watcher
 */
/**
 * @typedef { object } staticServer
 * @property { () => Promise<void> } destroy
 * @property { number } port
 */

const { destroyClients: destroyClients$1 } = pushEvents;
const { find: find$3, getProjectPath: getProjectPath$6 } = file;
const {
  handleMockResponse: handleMockResponse$2,
  handleMockWebSocket: handleMockWebSocket$2,
  handlePushEvent: handlePushEvent$2
} = common;
const { info: info$5, noisyInfo: noisyInfo$5 } = log;
const {
  isHtmlRequest: isHtmlRequest$2,
  isModuleBundlerFilePath: isModuleBundlerFilePath$3,
  isProjectFilePath: isProjectFilePath$2
} = is;
const { bundle: bundle$3 } = bundler_1;


const debug$8 = debug$a('dvlp:static');


const { interceptFileRead: interceptFileRead$2 } = intercept;

const { patchResponse: patchResponse$2 } = patch;






/**
 * Create static server
 *
 * @param { object } [options]
 * @param { string } [options.mockPath]
 * @param { number } [options.port]
 * @param { reloadServer } [options.reloader]
 * @param { object } [options.rollupConfig]
 * @param { (filePath: string, isServer: boolean) => Promise<string> | string | undefined } [options.transpiler]
 * @returns { staticServer }
 */
var staticServer = async function staticServer({
  mockPath,
  port,
  reloader,
  rollupConfig,
  transpiler
} = {}) {
  const server = new StaticServer(
    port,
    reloader,
    rollupConfig,
    transpiler,
    mockPath
  );

  await server.start();

  info$5(
    `\n  💥 Serving ${chalk.green(
      config.directories
        .map((dir) => getProjectPath$6(dir) || 'project')
        .join(', ')
    )} at ${chalk.green.underline(`http://localhost:${port}`)}`
  );

  return {
    port,
    destroy: server.destroy.bind(server)
  };
};

class StaticServer {
  /**
   * Constructor
   *
   * @param { number } port
   * @param { reloadServer } [reloader]
   * @param { object } [rollupConfig]
   * @param { (filePath: string, isServer: boolean) => Promise<string> | string | undefined } [transpiler]
   * @param { string } [mockPath]
   */
  constructor(port, reloader, rollupConfig, transpiler, mockPath) {
    this.lastChanged = '';
    this.mocks = mockPath && new mock(mockPath);
    this.patchResponseOptions = {
      rollupConfig,
      footerScript: {
        string: reloader && reloader.client
      },
      headerScript: {
        string: mockPath && this.mocks.client
      }
    };
    this.port = port;
    this.origin = `http://localhost:${port}`;
    this.reloader = reloader;
    this.rollupConfig = rollupConfig;
    this.server = null;
    this.transpiler = transpiler;
    this.transpilerCache = transpiler && new Map();

    if (reloader) {
      this.watcher = this.createWatcher();
      // Listen for all upcoming file system reads (including require('*'))
      this.unlistenForFileRead = interceptFileRead$2((filePath) => {
        if (isProjectFilePath$2(filePath)) {
          this.watcher.add(filePath);
        }
      });
    }
  }

  /**
   * Create watcher instance and listen for file changes
   *
   * @returns { watcher }
   */
  createWatcher() {
    return watch((filePath) => {
      noisyInfo$5(
        `\n  ⏱  ${new Date().toLocaleTimeString()} ${chalk.yellow(
          getProjectPath$6(filePath)
        )}`
      );

      this.lastChanged = filePath;

      // Will trigger one of:
      // 1. single css refresh if css and a link.href matches filePath
      // 2. multiple css refreshes if css and no link.href matches filePath (ie. it's a dependency)
      // 3. full page reload
      this.reloader.send(`/${path.relative(process.cwd(), filePath)}`);
    });
  }

  /**
   * Start server
   *
   * @returns { Promise<void> }
   */
  start() {
    return new Promise((resolve, reject) => {
      this.server = http.createServer(async (req, res) => {
        stopwatch.start(req.url);

        const url = req.url;
        let filePath;

        res.url = req.url;

        if (
          handleMockResponse$2(req, res, this.mocks) ||
          handlePushEvent$2(req, res, this.mocks)
        ) {
          return;
        }

        res.once('finish', () => {
          const duration = stopwatch.stop(url, true, true);
          const reroute =
            url !== req.url ? `(re-routed to ${chalk.green(req.url)})` : '';

          // 'transpiled' is added by transpile utility if handled there
          if (!res.transpiled) {
            info$5(
              res.statusCode < 300
                ? `${duration} handled request for ${chalk.green(
                    url
                  )} ${reroute}`
                : `${duration} [${
                    res.statusCode
                  }] unhandled request for ${chalk.red(url)} ${reroute}`
            );
          }
        });

        filePath = find$3(req);

        if (filePath) {
          if (isModuleBundlerFilePath$3(filePath)) {
            await bundle$3(path.basename(filePath), undefined, this.rollupConfig);
          }
        } else {
          // Re-write to root index.html
          if (isHtmlRequest$2(req)) {
            req.url = '/';
            filePath = find$3(req);
          }
        }

        if (!filePath) {
          debug$8(`not found "${req.url}"`);
          res.writeHead('404');
          return res.end();
        }

        patchResponse$2(filePath, req, res, this.patchResponseOptions);

        this.watcher && this.watcher.add(filePath);
        this.transpiler &&
          (await transpile(filePath, res, {
            transpilerCache: this.transpilerCache,
            lastChanged: this.lastChanged,
            transpiler: this.transpiler
          }));

        // Not transpiled
        if (!res.finished) {
          debug$8(`sending "${filePath}"`);
          return send(req, filePath, {
            cacheControl: true,
            dotfiles: 'allow',
            maxAge: config.maxAge
          }).pipe(res);
        }
      });

      this.server.keepAliveTimeout = 0;
      serverDestroy(this.server);

      this.server.on('error', reject);
      this.server.on('listening', resolve);
      this.server.on('upgrade', (req, socket, body) => {
        handleMockWebSocket$2(req, socket, body, this.mocks);
      });

      this.server.listen(this.port);
    });
  }

  /**
   * Stop running server
   *
   * @returns { Promise<void> }
   */
  stop() {
    return new Promise((resolve) => {
      if (!this.server) {
        return resolve();
      }

      this.server.removeAllListeners();
      this.server.destroy(() => {
        debug$8('server stopped');
        resolve();
      });
    });
  }

  /**
   * Destroy instance
   *
   * @returns { Promise<void> }
   */
  destroy() {
    debug$8('destroying');
    destroyClients$1();
    if (this.watcher) {
      this.unlistenForFileRead();
      this.watcher.close();
    }
    return this.stop();
  }
}

/**
 * @typedef { object } server
 * @property { () => Promise<void> } destroy
 */

const {
  exists,
  expandPath,
  getProjectPath: getProjectPath$7,
  importModule: importModule$1
} = file;

const { destroyWorkers: destroyWorkers$1 } = bundler_1;



const { info: info$6 } = log;




/**
 * Create server
 *
 * @param { string } filePath
 * @param { object } [options]
 * @param { string | Array<string> } [options.mockPath]
 * @param { number } [options.port]
 * @param { boolean } [options.reload]
 * @param { string } [options.transpiler]
 * @returns { server }
 */
var server = async function server(
  filePath = process.cwd(),
  { mockPath, port = config.port, reload = true, silent, transpiler } = {}
) {
  filePath = expandPath(filePath);
  mockPath = expandPath(mockPath);

  if (filePath.length === 1) {
    filePath = filePath[0];
  }

  exists(filePath);

  const isStatic =
    Array.isArray(filePath) ||
    fs.statSync(path.resolve(filePath)).isDirectory();
  let reloader$1, rollupConfig, server;

  config.directories.push(
    ...(Array.isArray(filePath) ? filePath : [filePath]).map((directory) => {
      directory = path.resolve(directory);
      if (fs.statSync(directory).isFile()) {
        directory = path.dirname(directory);
      }
      return directory;
    })
  );

  if (fs.existsSync(config.rollupConfigPath)) {
    rollupConfig = importModule$1(config.rollupConfigPath);
    info$6(
      `${chalk.green('✔')} registered custom Rollup.js config at ${chalk.green(
        getProjectPath$7(config.rollupConfigPath)
      )}`
    );
  }
  if (transpiler) {
    const transpilerpath = path.resolve(transpiler);

    transpiler = importModule$1(transpilerpath);
    info$6(
      `${chalk.green('✔')} loaded transpiler from ${chalk.green(
        getProjectPath$7(transpilerpath)
      )}`
    );
  }
  if (reload) {
    reloader$1 = await reloader();
  }
  if (isStatic) {
    server = await staticServer({
      mockPath,
      port,
      reloader: reloader$1,
      rollupConfig,
      transpiler
    });
  } else {
    server = await appServer(filePath, {
      mockPath,
      port,
      reloader: reloader$1,
      rollupConfig,
      transpiler
    });
  }

  config.activePort = server.port;

  info$6('👀 watching for changes...\n');

  if (silent) {
    log.silent = true;
  }

  return {
    destroy() {
      return Promise.all([
        reloader$1 && reloader$1.destroy(),
        server.destroy(),
        destroyWorkers$1()
      ]);
    }
  };
};

/**
 * @typedef { import("http").ClientRequest } ClientRequest
 * @typedef { import("http").ServerResponse } ServerResponse
 * @typedef { import("../mock/index.js").MockPushEvent } MockPushEvent
 * @typedef { import("../mock/index.js").MockPushStream } MockPushStream
 * @typedef { import("../mock/index.js").MockRequest } MockRequest
 * @typedef { import("../mock/index.js").MockResponse } MockResponse
 * @typedef { import("../push-events/index.js").PushEvent } PushEvent
 * @typedef { import("../push-events/index.js").PushStream } PushStream
 */
/**
 * @typedef { object } testServerOptions
 * @property { boolean } [autorespond] - automatically respond with dummy response when no matching file/mock found
 * @property { number } [latency] - minimum amount of random artificial latency to introduce (in `ms`) for responses
 * @property { number } [port] - port number
 * @property { string } [webroot] - the subpath from `process.cwd()` to prepend to relative paths
 */

const {
  connectClient: connectClient$2,
  destroyClients: destroyClients$2,
  pushEvent: pushEvent$2
} = pushEvents;

const debug$9 = debug$a('dvlp:test');



const { interceptClientRequest: interceptClientRequest$2 } = intercept;
const { isLocalhost: isLocalhost$2 } = is;



const { URL: URL$3 } = url$1;


const { EventSource: EventSource$3 } = fayeWebsocket;
const instances = new Set();
let reroute = false;
let networkDisabled = false;

interceptClientRequest$2((url) => {
  const isMocked = Array.from(instances).some((instance) =>
    instance.mocks.hasMatch(url)
  );
  let hostname = url.hostname || url.host;

  // Allow mocked requests to pass-through
  if (!isMocked && !isLocalhost$2(hostname)) {
    if (reroute) {
      // Reroute back to this server
      url.host = url.hostname = `localhost:${config.activePort}`;
    } else if (networkDisabled) {
      throw Error(`network connections disabled. Unable to request ${url}`);
    }
  }

  return url;
});

/**
 * Create test server
 *
 * @param { testServerOptions } [options]
 * @returns { TestServer }
 */
var testServer = async function testServer({
  autorespond = true,
  latency = config.latency,
  port = config.port,
  webroot = ''
} = {}) {
  const server = new TestServer({
    autorespond,
    latency,
    port,
    webroot
  });

  await server._start();
  // Make sure 'mock' has access to current active port
  config.activePort = port;

  instances.add(server);

  return server;
};

/**
 * Disable all external network connections
 * and optionally reroute all external requests to this server
 *
 * @param { boolean } [rerouteAllRequests]
 * @returns { void }
 */
var disableNetwork = function disableNetwork(
  rerouteAllRequests = false
) {
  networkDisabled = true;
  reroute = rerouteAllRequests;
};

/**
 * Enable all external network connections
 *
 * @returns { void }
 */
var enableNetwork = function enableNetwork() {
  networkDisabled = true;
  reroute = false;
};

class TestServer {
  /**
   * Constructor
   *
   * @param { testServerOptions } options
   */
  constructor(options) {
    const { autorespond, latency, port, webroot } = options;

    this.latency = latency;
    this.mocks = new mock();
    this.webroot = webroot;
    this._autorespond = autorespond;
    this._port = port;
    this._server = null;
  }

  /**
   * Start server
   *
   * @returns { Promise<void> }
   * @private
   */
  _start() {
    return new Promise((resolve, reject) => {
      this._server = http.createServer(async (req, res) => {
        if (EventSource$3.isEventSource(req)) {
          connectClient$2(
            {
              url: req.url,
              type: 'es'
            },
            req,
            res
          );
          this.pushEvent(req.url, 'connect');
          return;
        }

        const url = new URL$3(req.url, `http://localhost:${this._port}`);
        const error = url.searchParams.get('error') != null;
        const hang = url.searchParams.get('hang') != null;
        const maxage = url.searchParams.get('maxage') || 0;
        const missing = url.searchParams.get('missing') != null;
        const offline = url.searchParams.get('offline') != null;
        const mock = url.searchParams.get('dvlpmock');

        if (hang) {
          return;
        }

        this.latency && (await sleep$1(this.latency));

        if (error || missing) {
          const statusCode = error ? 500 : 404;
          const body = error ? 'error' : 'missing';

          debug$9(`not ok: ${req.url} responding with ${statusCode}`);
          res.statusCode = statusCode;
          res.end(body);
          return;
        } else if (offline) {
          debug$9(`not ok: ${req.url} offline`);
          req.socket.destroy();
          return;
        } else if (req.aborted) {
          debug$9(`not ok: ${req.url} aborted`);
          return;
        } else if (mock) {
          debug$9(`ok: ${req.url} responding with mocked data`);
          this.mocks.matchResponse(mock, res);
          return;
        }

        const trimmedPath = url.pathname.slice(1);
        let type = mime.getType(trimmedPath);
        // TODO: handle encoded query strings in path name?
        let filePath = path.resolve(path.join(this.webroot, trimmedPath));
        let body = '';
        let headers = {};
        let size = 0;
        let stat;
        let msg = '';

        // Ignore webroot if no file
        if (!fs.existsSync(filePath)) {
          filePath = path.resolve(trimmedPath);
        }

        try {
          stat = fs.statSync(filePath);
          size = stat.size;
          msg = `ok: ${req.url} responding with file`;
        } catch (err) {
          if (!this._autorespond) {
            res.writeHead(404);
            return res.end();
          }
          body = '"hello"';
          size = Buffer.byteLength(body);
          msg = `ok: ${req.url} responding with dummy file`;
        }

        res.writeHead(200, {
          'Content-Length': size,
          'Cache-Control': `public, max-age=${maxage}`,
          'Content-Type': type,
          ...headers
        });

        debug$9(msg);

        return body ? res.end(body) : fs.createReadStream(filePath).pipe(res);
      });

      serverDestroy(this._server);
      this._server.unref();
      this._server.on('error', reject);
      this._server.on('listening', resolve);
      this._server.on('upgrade', (req, socket, body) => {
        if (fayeWebsocket.isWebSocket(req)) {
          connectClient$2(
            {
              url: req.url,
              type: 'ws'
            },
            req,
            socket,
            body
          );
          this.pushEvent(req.url, 'connect');
        }
      });

      this._server.listen(this._port);
    });
  }

  /**
   * Load mock files at 'filePath'
   *
   * @param { string | Array<string> } filePath
   */
  loadMockFiles(filePath) {
    this.mocks.load(filePath);
  }

  /**
   * Register mock 'response' for 'request'
   *
   * @param { string | MockRequest } request
   * @param { MockResponse } response
   * @param { boolean } [once]
   */
  mockResponse(request, response, once = false) {
    this.mocks.addResponse(request, response, once);
  }

  /**
   * Register mock push 'events' for 'stream'
   *
   * @param { string | MockPushStream } stream
   * @param { MockPushEvent | Array<MockPushEvent> } events
   */
  mockPushEvents(stream, events) {
    this.mocks.addPushEvents(stream, events);
  }

  /**
   * Push data to WebSocket/EventSource clients
   * A string passed as 'event' will be handled as a named mock push event
   *
   * @param { string | PushStream } stream
   * @param { string | PushEvent } [event]
   * @returns { void }
   */
  pushEvent(stream, event) {
    // Passed a mocked event name
    if (typeof event === 'string') {
      this.mocks.matchPushEvent(stream, event, pushEvent$2);
    } else {
      pushEvent$2(stream, event);
    }
  }

  /**
   * Stop running server
   *
   * @returns { Promise<void> }
   * @private
   */
  _stop() {
    return new Promise((resolve) => {
      if (!this._server) {
        return resolve();
      }

      this._server.removeAllListeners();
      this._server.destroy(() => {
        debug$9('server stopped');
        resolve();
      });
    });
  }

  /**
   * Destroy instance
   *
   * @returns { Promise<void> }
   */
  destroy() {
    debug$9('destroying');
    destroyClients$2();
    this.mocks.clean();
    instances.delete(this);
    return this._stop();
  }
}

/**
 * Sleep for random number of milliseconds between 'min' and '2xmin'
 *
 * @param { number } min
 * @returns { Promise<void> }
 * @private
 */
function sleep$1(min) {
  return new Promise((resolve) => {
    if (!min) {
      return resolve();
    }
    setTimeout(resolve, min + Math.random() * min);
  });
}
testServer.disableNetwork = disableNetwork;
testServer.enableNetwork = enableNetwork;

var server$1 = server;
var testServer$1 = testServer;

var lib = {
	server: server$1,
	testServer: testServer$1
};

exports.default = lib;
exports.server = server$1;
exports.testServer = testServer$1;
